import sys
from logging import Logger
from pathlib import Path
from pypomes_core import exc_format, str_is_float
from typing import Any, BinaryIO, Literal

from .db_common import (
    _DB_CONN_DATA, _DB_LOGGERS,
    _BUILTIN_FUNCTIONS, DB_BIND_META_TAG, DbEngine, DbParam,
    _assert_engine, _get_param, _bind_columns, _bind_marks,
    _combine_insert_data, _combine_update_data, _combine_search_data
)
from .pool_pomes import db_pool_acquire, db_pool_release


def db_setup(engine: DbEngine,
             db_name: str,
             db_user: str,
             db_pwd: str,
             db_host: str,
             db_port: int,
             db_client: str | Path = None,
             db_driver: str = None) -> bool:
    """
    Establish the parameters for access to *engine*.

    The meaning of some parameters may vary between different database engines.
    All parameters, are required, with these exceptions:
        - *db_client* may be provided for *oracle*, only
        - *db_driver* is required for *sqlserver*, only

    :param engine: the database engine (one of [*mysql*, *oracle*, *postgres*, *sqlserver*])
    :param db_name: the database or service name
    :param db_user: the logon user
    :param db_pwd: the logon password
    :param db_host: the host URL
    :param db_port: the connection port (a positive integer)
    :param db_driver: the database driver (SQLServer only)
    :param db_client: the path to the client software (Oracle only)
    :return: *True* if the data was accepted, *False* otherwise
    """
    # initialize the return variable
    result: bool = False

    if engine in DbEngine and engine != DbEngine.SPANNER and \
       db_name and db_user and db_pwd and db_host and \
       not (engine != DbEngine.ORACLE and db_client) and \
       not (engine != DbEngine.SQLSERVER and db_driver) and \
       not (engine == DbEngine.SQLSERVER and not db_driver) and \
       isinstance(db_port, int) and db_port > 0:
        _DB_CONN_DATA[engine] = {
            DbParam.NAME: db_name,
            DbParam.USER: db_user,
            DbParam.PWD: db_pwd,
            DbParam.HOST: db_host,
            DbParam.PORT: db_port,
            DbParam.VERSION: ""
        }
        if engine == DbEngine.ORACLE:
            _DB_CONN_DATA[engine][DbParam.CLIENT] = Path(db_client)
        elif engine == DbEngine.SQLSERVER:
            _DB_CONN_DATA[engine][DbParam.DRIVER] = db_driver
        result = True

    return result


def db_set_logger(engine: DbEngine = None,
                  logger: Logger = None) -> bool:
    """
    Establish the logger for logging operations involving *engine*.

    This operation must be invoked after *engine* has been configured. If specified,
    *engine* must be one of [*mysql*, *oracle*, *postgres*, *sqlserver*, *spanner*].

    :param engine: the database engine to use (uses the default engine, if not provided)
    :param logger: the logger for the database engine
    :return: *True* if the logger was estabished, *False* otherwise
    """
    # initialize the return variable
    result: bool = False

    # assert the database engine
    engine = _assert_engine(engine=engine)
    if engine:
        _DB_LOGGERS[engine] = logger
        result = True

    return result


def db_assert_access(engine: DbEngine = None,
                     errors: list[str] = None) -> bool:
    """
    Determine whether the *engine*'s current configuration allows for connections.

    This function should be invoked once, at application's initialization time. This is necessary
    to make sure connections are obtainable with the provided parameters, to establish the version
    of the database engine in use, and, if applicable in the case of *Oracle*, to initialize access
    to its client library.

    :param engine: the database engine to use (uses the default engine, if not provided)
    :param errors: incidental error messages (might be a non-empty list)
    :return: *True* if the access attempt succeeded, *False* otherwise
    """
    # initialize the return variable
    result: bool = False

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    if engine in _DB_CONN_DATA:
        # determine if access to 'engine' has been asserted
        version: str = _DB_CONN_DATA[engine].get(DbParam.VERSION)
        if version:
            result = True
        elif engine == DbEngine.SPANNER:
            from .spanner_frame import SpannerParam, GoogleSpanner
            spanner: GoogleSpanner = _DB_CONN_DATA[engine][SpannerParam.ENGINE]
            result = spanner.initialize(errors=errors)
        else:
            # assert access to 'engine'
            if engine == DbEngine.ORACLE:
                from . import oracle_pomes
                oracle_pomes.initialize(errors=errors)
            conn: Any = db_connect(engine=engine,
                                   errors=errors)
            if conn:
                # noinspection PyTypeChecker
                _DB_CONN_DATA[engine][DbParam.VERSION] = __get_version(engine=engine,
                                                                       connection=conn)
                db_close(connection=conn,
                         engine=engine)
                result = True

    return result


def db_get_default_engine() -> DbEngine | None:
    """
    Retrieve the database engine configred as the default engine.

    The default engine is typically the first database engine configured, either from environment variables or
    by an invocation of *db_setup()*.

    :return: the default database engine, or *None* if no engine has been configured
    """
    return next(iter(_DB_CONN_DATA)) if _DB_CONN_DATA else None


def db_get_engines() -> list[DbEngine]:
    """
    Retrieve the *list* of configured engines.

    This *list* may include any of the supported engines:
    *mysql*, *oracle*, *postgres*, *sqlserver*, *spanner*.
    Note that the values in the returned *list* are instances of *DbEngine*, not strings.

    :return: the *list* of configured engines
    """
    return list(_DB_CONN_DATA)


def db_get_param(key: DbParam,
                 engine: DbEngine = None) -> Any:
    """
    Return the current value for connection parameter *key*.

    The *key* should be one of *name*, *user*, *pwd*, *host*, and *port*.
    For *oracle* and *sqlserver* engines, the extra keys *client* and *driver* may be used, respectively.
    Not applicable to *spanner* engines.

    :param key: the reference parameter
    :param engine: the reference database engine (the default engine, if not provided)
    :return: the current value of the connection parameter, or *None* if not found
    """
    # initialize the return variable
    result: Any = None

    # assert the database engine
    engine: DbEngine = next(iter(_DB_CONN_DATA)) if not engine and _DB_CONN_DATA else engine

    # retrieve the connection parameter
    if engine != DbEngine.SPANNER:
        result = _get_param(engine=engine,
                            param=key)
    return result


def db_get_params(engine: DbEngine = None) -> dict[DbParam, Any] | None:
    """
    Return the current connection parameters for *engine* as a *dict*.

    The returned *dict* contains the keys *name*, *user*, *pwd*, *host*, and *port*.
    For *oracle* engines, the returned *dict* contains the extra key *client*.
    For *sqlserver* engines, the returned *dict* contains the extra key *driver*.
    The meaning of these parameters may vary between different database engines.
    Not applicable to *spanner* engines.

    :param engine: the reference database engine (the default engine, if not provided)
    :return: the current connection parameters for *engine*, or *None* if not found
    """
    # initialize the return variable
    result: dict[DbParam, Any] | None = None

    # assert the database engine
    engine: DbEngine = next(iter(_DB_CONN_DATA)) if not engine and _DB_CONN_DATA else engine

    # retrieve the connection parameters
    if engine in _DB_CONN_DATA and engine != DbEngine.SPANNER:
        result = _DB_CONN_DATA[engine].copy()

    return result


def db_get_connection_string(engine: DbEngine = None) -> str:
    """
    Build and return the connection string for connecting to the database.

    Not applicable to *spanner* engines.

    :param engine: the reference database engine (the default engine, if not provided)
    :return: the connection string, or *None* if error
    """
    # initialize the return variable
    result: str | None = None

    # assert the database engine
    engine = next(iter(_DB_CONN_DATA)) if not engine and _DB_CONN_DATA else engine

    if engine == DbEngine.MYSQL:
        pass
    elif engine == DbEngine.ORACLE:
        from . import oracle_pomes
        result = oracle_pomes.get_connection_string()
    elif engine == DbEngine.POSTGRES:
        from . import postgres_pomes
        result = postgres_pomes.get_connection_string()
    elif engine == DbEngine.SQLSERVER:
        from . import sqlserver_pomes
        result = sqlserver_pomes.get_connection_string()

    return result


def db_get_reserved_words(engine: DbEngine = None) -> list[str] | None:
    """
    Obtain and return the list of reserved words for *engine*.

    Reserved words can not be used to name database objects like tables, columns, triggers, constraints, etc.

    :param engine: the reference database engine (the default engine, if not provided)
    :return: the engine's list of reserved words, or *None* if error
    """
    # initialize the return variable
    result: list[str] | None = None

    # assert the database engine
    engine = next(iter(_DB_CONN_DATA)) if not engine and _DB_CONN_DATA else engine

    match engine:
        case DbEngine.MYSQL:
            from . import mysql_pomes
            result = mysql_pomes.RESERVED_WORDS
        case DbEngine.ORACLE:
            from . import oracle_pomes
            result = oracle_pomes.RESERVED_WORDS
        case DbEngine.POSTGRES:
            from . import postgres_pomes
            result = postgres_pomes.RESERVED_WORDS
        case DbEngine.SQLSERVER:
            from . import sqlserver_pomes
            result = sqlserver_pomes.RESERVED_WORDS

    return result


def db_is_reserved_word(word: str,
                        engine: DbEngine = None) -> bool | None:
    """
    Verify whether *word* is a reserved word for *engine*.

    Reserved words can not be used to name database objects like tables, columns, triggers, constraints, etc.

    :param word: the word to verify
    :param engine: the reference database engine (the default engine, if not provided)
    :return: *True* if *word* is a reserved word for *engine*, *False* otherwise, *None* on error
    """
    # initialize the return variable
    result: bool | None = None

    # assert the database engine
    engine = next(iter(_DB_CONN_DATA)) if not engine and _DB_CONN_DATA else engine

    if isinstance(word, str):
        word = word.upper()
    match engine:
        case DbEngine.MYSQL:
            from . import mysql_pomes
            result = word in mysql_pomes.RESERVED_WORDS
        case DbEngine.ORACLE:
            from . import oracle_pomes
            result = word in oracle_pomes.RESERVED_WORDS
        case DbEngine.POSTGRES:
            from . import postgres_pomes
            result = word in postgres_pomes.RESERVED_WORDS
        case DbEngine.SQLSERVER:
            from . import sqlserver_pomes
            result = word in sqlserver_pomes.RESERVED_WORDS

    return result


def db_adjust_placeholders(stmt: str,
                           engine: DbEngine = None) -> str:
    """
    Replace the occurrences of bind meta-tag in *stmt*, with the appropriate placeholder for *engine*.

    The bind meta-tag is defined by *DB_BIND_META_TAG*, an environment variable with the default value *%?*.
    These are the placeholders specific to the supported DB engines:
        - mysql:     ?
        - oracle:    :<n> (1-based)
        - postgres:  %s
        - sqlserver: ?
        - spanner:   @gs<n> (1-based)

    :param stmt: the statement for which to replace the bind meta-tags with the proper placeholders
    :param engine: the reference database engine (the default engine, if not provided)
    :return: the statement with the proper placeholders, or *None* if the engine is not known
    """
    # initialize the return variable
    result: str | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine)

    # adjust the placeholders
    match engine:
        case DbEngine.POSTGRES:
            result = stmt.replace(DB_BIND_META_TAG, "%s")
        case DbEngine.ORACLE:
            pos: int = 0
            result = stmt
            while result.find(DB_BIND_META_TAG) > 0:
                pos += 1
                result = result.replace(DB_BIND_META_TAG, f":{pos}", 1)
        case DbEngine.MYSQL | DbEngine.SQLSERVER:
            result = stmt.replace(DB_BIND_META_TAG, "?")
        case DbEngine.SPANNER:
            pos: int = 0
            result = stmt
            while result.find(DB_BIND_META_TAG) > 0:
                pos += 1
                result = result.replace(DB_BIND_META_TAG, f"@gs{pos}", 1)

    return result


def db_bind_arguments(stmt: str,
                      bind_vals: list[Any],
                      engine: DbEngine = None) -> str:
    """
    Replace the placeholders in *query_stmt* with the values in *bind_vals*, and return the modified query statement.

    The placeholders in *query_stmt* can be either specific to *engine*, or the generic value set in
    DB_BIND_META_TAG (default value is '%?').
    These are the placeholders specific to the supported DB engines:
        - mysql:     ?
        - oracle:    :n (1-based)
        - postgres:  %s
        - sqlserver: ?

    Note that using a statement in a situation where values for types other than *bool*, *str*, *int*, *float*,
    *date*, or *datetime* were replaced, may bring about undesirable consequences, as the standard string
    representations for these other types would be used.

    :param stmt: the query statement
    :param bind_vals: the values to replace the placeholders with
    :param engine: the database engine to use (uses the default engine, if not provided)
    :return: the query statement with the placeholders replaced with their corresponding values
    """
    # initialize the return variable
    result: str | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine)

    # establish the correct bind tags
    stmt = db_adjust_placeholders(stmt=stmt,
                                  engine=engine)
    # bind the arguments
    if engine == DbEngine.MYSQL:
        from . import mysql_pomes
        result = mysql_pomes.bind_arguments(stmt=stmt,
                                            bind_vals=bind_vals)
    elif engine == DbEngine.ORACLE:
        from . import oracle_pomes
        result = oracle_pomes.bind_arguments(stmt=stmt,
                                             bind_vals=bind_vals)
    elif engine == DbEngine.POSTGRES:
        from . import postgres_pomes
        result = postgres_pomes.bind_arguments(stmt=stmt,
                                               bind_vals=bind_vals)
    elif engine == DbEngine.SQLSERVER:
        from . import sqlserver_pomes
        result = sqlserver_pomes.bind_arguments(stmt=stmt,
                                                bind_vals=bind_vals)
    return result


def db_convert_default(value: str,
                       source_engine: DbEngine,
                       target_engine: DbEngine) -> str | None:
    """
    Convert the default value used in *source_engine* to its equivalent in *target_engine*.

    :param value: the value to be converted
    :param source_engine: the source database engine
    :param target_engine: the target database engine
    :return: the converted value, or *None* if no convertion was possible.
    """
    # initialize the return variable
    result: str | None = None

    # note that 'str_is_int()' is not necessary here
    if str_is_float(value) or (value.startswith("'") and value.endswith("'")):
        # 'value' is a literal (numeric or string)
        result = value
    else:
        pos_source: int | None = None
        match source_engine:
            case DbEngine.MYSQL:
                pos_source = 0
            case DbEngine.ORACLE:
                pos_source = 1
            case DbEngine.POSTGRES:
                pos_source = 2
            case DbEngine.SQLSERVER:
                pos_source = 3

        pos_target: int | None = None
        match target_engine:
            case DbEngine.MYSQL:
                pos_target = 0
            case DbEngine.ORACLE:
                pos_target = 1
            case DbEngine.POSTGRES:
                pos_target = 2
            case DbEngine.SQLSERVER:
                pos_target = 3

        for func in _BUILTIN_FUNCTIONS:
            if func[pos_source] == value.upper():
                result = func[pos_target]
                break

    return result


def db_connect(autocommit: bool = False,
               engine: DbEngine = None,
               errors: list[str] = None) -> Any:
    """
    Obtain and return a connection to the database, or *None* if the connection cannot be obtained.

    The target database engine, specified or default, must have been previously configured.
    Obtaining a connection from the connection pool is initially attempted, and if not successful,
    the native driver is summoned for the task.

    :param autocommit: whether the connection is to be in autocommit mode (defaults to False)
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param errors: incidental error messages (might be a non-empty list)
    :return: the connection to the database
    """
    # initialize the return variable
    result: int | None = None

    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)
    if engine:
        # attempt to obtain a connection from the pool
        result: Any = db_pool_acquire(engine=engine)
        if not result:
            # connect to the database
            if engine == DbEngine.MYSQL:
                pass
            elif engine == DbEngine.ORACLE:
                from . import oracle_pomes
                result = oracle_pomes.connect(autocommit=autocommit,
                                              errors=curr_errors)
            elif engine == DbEngine.POSTGRES:
                from . import postgres_pomes
                result = postgres_pomes.connect(autocommit=autocommit,
                                                errors=curr_errors)
            elif engine == DbEngine.SQLSERVER:
                from . import sqlserver_pomes
                result = sqlserver_pomes.connect(autocommit=autocommit,
                                                 errors=curr_errors)
    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def db_commit(connection: Any,
              engine: DbEngine = None,
              errors: list[str] = None) -> None:
    """
    Commit the current transaction on *connection*.

    :param connection: the reference database connection
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param errors: incidental error messages (might be a non-empty list)
    """
    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    # commit the transaction
    try:
        connection.commit()
        if engine and _DB_LOGGERS[engine]:
            _DB_LOGGERS[engine].debug(f"Transaction committed on '{id(connection)}'")
    except Exception as e:
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if engine and _DB_LOGGERS[engine]:
            _DB_LOGGERS[engine].error(msg=exc_err)
        if isinstance(errors, list):
            errors.append(exc_err)


def db_rollback(connection: Any,
                engine: DbEngine = None) -> None:
    """
    Rollback the current transaction on *connection*.

    :param connection: the reference database connection
    :param engine: the database engine to use (uses the default engine, if not provided)
    """
    # assert the database engine
    engine = _assert_engine(engine=engine)

    # rollback the transaction
    try:
        connection.rollback()
        if engine and _DB_LOGGERS[engine]:
            _DB_LOGGERS[engine].debug(f"Transaction rolled back on '{id(connection)}'")
    except Exception as e:
        if engine and _DB_LOGGERS[engine]:
            exc_err: str = exc_format(exc=e,
                                      exc_info=sys.exc_info())
            _DB_LOGGERS[engine].error(msg=exc_err)


def db_close(connection: Any,
             engine: DbEngine = None) -> None:
    """
    Close the connection given in *connection*.

    Returning *connection* to the connection pool is initially attempted, and if not successful,
    a *close()* operation is performed on the connection itself.

    :param connection: the reference database connection
    :param engine: the database engine to use (uses the default engine, if not provided)
    """
    # assert the database engine
    engine = _assert_engine(engine=engine)

    # attempt to return the connection to the pool
    if not db_pool_release(connection=connection,
                           engine=engine):
        # close the connection
        try:
            connection.close()
            if engine and _DB_LOGGERS[engine]:
                _DB_LOGGERS[engine].debug(f"Connection {id(connection)} closed")
        except Exception as e:
            if engine and _DB_LOGGERS[engine]:
                exc_err: str = exc_format(exc=e,
                                          exc_info=sys.exc_info())
                _DB_LOGGERS[engine].error(msg=exc_err)


def db_add_query_limits(sel_stmt: str,
                        offset_count: int | None,
                        limit_count: int | None) -> str:
    """
    Add offset and limit to *sel_stmt* according to the *SQL* standard..

    If not positive integers, *offset_count*, and *limit_count* are ignored. The parameter *offset_count*
    is used to offset the retrieval of tuples, whereas *limit_count* establishes a ceiling on the number
    of tuples returned.

    :param sel_stmt: the query statement to modify
    :param offset_count: number of tuples to skip
    :param limit_count: limit to the number of tuples returned
    :return: the modified query statement
    """
    # initialize the return variable
    result: str = sel_stmt

    # establish a limit to the number of tuples returned
    if isinstance(limit_count, int) and limit_count > 0:
        result += f" LIMIT {limit_count}"

    # establish an offset into the result set
    if isinstance(offset_count, int) and offset_count > 0:
        result += f" OFFSET {offset_count}"

    return result


def db_build_stmt(base_stmt: str,
                  modify_vals: tuple = None,
                  modify_data: dict[str, Any] = None,
                  where_clause: str = None,
                  where_vals: tuple = None,
                  where_data: dict[str | tuple |
                                   tuple[str,
                                         Literal["=", ">", "<", ">=", "<=", "<>",
                                                 "in", "like", "ilike", "between"] | None,
                                         Literal["and", "or"] | None], Any] = None,
                  orderby_clause: str = None,
                  offset_count: int = None,
                  limit_count: int = None,
                  engine: DbEngine = None,
                  errors: list[str] = None) -> str | None:
    """
    Build and return the *SQL* query statement as per the parameters provided.

    It is left to the invoker to make sure *base_stmt* makes sense and is consistent with the parameters
    provided:
        - it can refer to *SELECT*, *INSERT*, *UPDATE*, or *DELETE* operations
        - *modify_vals* and *modify_data* may be specified for *INSERT*s and *UPDATE*s only
        - *orderby_clause*, *offset_count*, and *limit_count* may be specified for *SELECT*s only
        _ *where_clause*, *where_vals*, and *where_data* may not be specified for *INSERTs*

    If applicable, selection criteria may be specified in *where_clause* and *where_vals*, or additionally
    but preferably, by key-value pairs in *where_data*. The syntax specific to *where_data*'s key/value pairs
    is as follows:
        1. *key*:
            - an attribute (possibly aliased), or
            - a 2/3-tuple with an attribute and the corresponding SQL comparison operation
              ("=", ">", "<", ">=", "<=", "<>", "in", "like", "ilike", "between" - defaults to "="), followed
              by a SQL logical operator relating it to the next item ("and", "or" - defaults to "and")
        2. *value*:
            - a scalar, or a list, or an expression possibly containing other attribute(s)

    :param base_stmt: the base *SQL* statement to build upon
    :param modify_vals: values for *INSERT* or *UPDATE* operations
    :param modify_data: insert or update data as key-value pairs
    :param where_clause: criteria for tuple selection (ignored if *sel_stmt* contains a *WHERE* clause)
    :param where_vals: values to be associated with the selection criteria
    :param where_data: selection criteria specified as key-value pairs
    :param orderby_clause: retrieval order (ignored if *base_stmt* contains a *ORDER BY* clause)
    :param offset_count: number of tuples to skip (defaults to none)
    :param limit_count: limit to the number of tuples returned, to be specified in the query statement itself
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param errors: incidental error messages (might be a non-empty list)
    :return: the *SQL* query statement built as per the given parameters , or *None* if error
    """
    # initialize the return variable
    result: str | None = None

    # determine the type of query
    is_select: bool = base_stmt.lower().startswith("select")
    is_insert: bool = base_stmt.lower().startswith("insert")
    is_update: bool = base_stmt.lower().startswith("update")
    is_delete: bool = base_stmt.lower().startswith("delete")

    # make sure parameters are consistent
    if (is_select or is_insert or is_update or is_delete) and \
       not (not is_select and (orderby_clause or offset_count or limit_count)) and \
       not (is_insert and (where_clause or where_vals or where_data)) and \
       not ((is_select or is_delete) and (modify_vals or modify_data)):

        # assert the database engine
        engine = _assert_engine(engine=engine,
                                errors=errors)
        if engine:
            result = base_stmt

            # process update data
            if modify_data:
                if is_insert:
                    result, modify_vals = _combine_insert_data(insert_stmt=result,
                                                               insert_vals=modify_vals,
                                                               insert_data=modify_data)
                else:
                    result, modify_vals = _combine_update_data(update_stmt=result,
                                                               update_vals=modify_vals,
                                                               update_data=modify_data)
            # process search data
            if where_clause or where_data or orderby_clause:
                result, where_vals = _combine_search_data(query_stmt=result,
                                                          where_clause=where_clause,
                                                          where_vals=where_vals,
                                                          where_data=where_data,
                                                          orderby_clause=orderby_clause,
                                                          engine=engine)
            # bind the arguments
            bind_vals: list = []
            if modify_vals:
                bind_vals.extend(list(modify_vals))
            if where_vals:
                bind_vals.extend(list(where_vals))
            if bind_vals:
                result = db_bind_arguments(stmt=result,
                                           bind_vals=bind_vals,
                                           engine=engine)
            # add limits
            if offset_count or limit_count:
                if engine == DbEngine.MYSQL:
                    pass
                elif engine == DbEngine.ORACLE:
                    from . import oracle_pomes
                    result = oracle_pomes.add_query_limits(sel_stmt=result,
                                                           offset_count=offset_count,
                                                           limit_count=limit_count)
                elif engine == DbEngine.POSTGRES:
                    result = db_add_query_limits(sel_stmt=result,
                                                 offset_count=offset_count,
                                                 limit_count=limit_count)
                elif engine == DbEngine.SQLSERVER:
                    from . import sqlserver_pomes
                    result = sqlserver_pomes.add_query_limits(sel_stmt=result,
                                                              offset_count=offset_count,
                                                              limit_count=limit_count)
    elif isinstance(errors, list):
        errors.append("Inconsistent parameters")

    return result


def db_count(table: str,
             count_clause: str = None,
             where_clause: str = None,
             where_vals: tuple = None,
             where_data: dict[str | tuple |
                              tuple[str,
                                    Literal["=", ">", "<", ">=", "<=", "<>",
                                            "in", "like", "ilike", "between"] | None,
                                    Literal["and", "or"] | None], Any] = None,
             engine: DbEngine = None,
             connection: Any = None,
             committable: bool = None,
             errors: list[str] = None) -> int | None:
    """
    Obtain and return the number of tuples in *table*, meeting the criteria provided.

    Optional selection criteria may be specified in *where_clause* and *where_vals*, or additionally but
    preferably, by key-value pairs in *where_data*. Care should be exercised if *where_clause* contains *IN*
    directives. In PostgreSQL, the list of values for an attribute with the *IN* directive must be contained
    in a specific tuple, and the operation will break for a list of values containing only 1 element.
    The safe way to specify *IN* directives is to add them to *where_data*, as the specifics of each DB flavor
    will then be properly dealt with.

    The syntax specific to *where_data*'s key/value pairs is as follows:
        1. *key*:
            - an attribute (possibly aliased), or
            - a 2/3-tuple with an attribute, the corresponding SQL comparison operation
              ("=", ">", "<", ">=", "<=", "<>", "in", "like", "ilike", "between" - defaults to "="), and
              a SQL logical operator relating it to the next item ("and", "or" - defaults to "and")
        2. *value*:
            - a scalar, or a list, or an expression possibly containing other attribute(s)

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param table: the table to be searched
    :param count_clause: optional parameters in the *COUNT* clause (defaults to 'COUNT(*)')
    :param where_clause: optional criteria for tuple selection
    :param where_vals: values to be associated with the selection criteria
    :param where_data: the selection criteria, specified as key-value pairs
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the number of tuples counted, or *None* if error
    """
    # initialize the return variable
    result: int | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # execute the query
    recs: list[tuple[int]] = db_select(sel_stmt=f"SELECT COUNT({count_clause or '*'}) FROM {table}",
                                       where_clause=where_clause,
                                       where_vals=where_vals,
                                       where_data=where_data,
                                       engine=engine,
                                       connection=connection,
                                       committable=committable,
                                       errors=errors)
    if isinstance(recs, list):
        result = recs[0][0]

    return result


def db_exists(table: str,
              where_clause: str = None,
              where_vals: tuple = None,
              where_data: dict[str | tuple |
                               tuple[str,
                                     Literal["=", ">", "<", ">=", "<=", "<>",
                                             "in", "like", "ilike", "between"] | None,
                                     Literal["and", "or"] | None], Any] = None,
              min_count: int = None,
              max_count: int = None,
              engine: DbEngine = None,
              connection: Any = None,
              committable: bool = None,
              errors: list[str] = None) -> bool | None:
    """
    Determine whether at least one tuple in *table* meets the criteria provided.

    Optional selection criteria may be specified in *where_clause* and *where_vals*, or additionally but
    preferably, by key-value pairs in *where_data*. Care should be exercised if *where_clause* contains *IN*
    directives. In PostgreSQL, the list of values for an attribute with the *IN* directive must be contained
    in a specific tuple, and the operation will break for a list of values containing only 1 element.
    The safe way to specify *IN* directives is to add them to *where_data*, as the specifics of each DB flavor
    will then be properly dealt with.

    The syntax specific to *where_data*'s key/value pairs is as follows:
        1. *key*:
            - an attribute (possibly aliased), or
            - a 2/3-tuple with an attribute, the corresponding SQL comparison operation
              ("=", ">", "<", ">=", "<=", "<>", "in", "like", "ilike", "between" - defaults to "="), and
              a SQL logical operator relating it to the next item ("and", "or" - defaults to "and")
        2. *value*:
            - a scalar, or a list, or an expression possibly containing other attribute(s)

    If not positive integers, *min_count* and *max_count* are ignored. If both *min_count* and *max_count*
    are specified with equal values, then exactly that number of tuples must exist.  If neither one is
    specified, than at least one tuple is expected to exist.

    The targer database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param table: the, possibly schema-qualified, name of the table to be searched
    :param where_clause: optional criteria for tuple selection
    :param where_vals: values to be associated with the selection criteria
    :param where_data: the selection criteria, specified as key-value pairs
    :param min_count: optionally defines the minimum number of tuples expected to exist
    :param max_count: optionally defines the maximum number of tuples expected to exist
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: *True* if the criteria for tuple existence were met, *False* otherwise, or *None* if error
    """
    # initialize the return variable
    result: bool | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # count the tuples
    count: int = db_count(table=table,
                          where_clause=where_clause,
                          where_vals=where_vals,
                          where_data=where_data,
                          engine=engine,
                          connection=connection,
                          committable=committable,
                          errors=errors)
    if isinstance(count, int):
        result = not (count == 0 or
                      (isinstance(max_count, int) and 0 < max_count < count) or
                      (isinstance(min_count, int) and min_count > 0 and min_count > count))
    return result


def db_select(sel_stmt: str,
              where_clause: str = None,
              where_vals: tuple = None,
              where_data: dict[str | tuple |
                               tuple[str,
                                     Literal["=", ">", "<", ">=", "<=", "<>",
                                             "in", "like", "ilike", "between"] | None,
                                     Literal["and", "or"] | None], Any] = None,
              orderby_clause: str = None,
              min_count: int = None,
              max_count: int = None,
              offset_count: int = None,
              limit_count: int = None,
              engine: DbEngine = None,
              connection: Any = None,
              committable: bool = None,
              errors: list[str] = None) -> list[tuple] | None:
    """
    Query the database and return all tuples that satisfy the criteria provided.

    Optional selection criteria may be specified in *where_clause* and *where_vals*, or additionally but
    preferably, by key-value pairs in *where_data*. Care should be exercised if *where_clause* contains *IN*
    directives. In PostgreSQL, the list of values for an attribute with the *IN* directive must be contained
    in a specific tuple, and the operation will break for a list of values containing only 1 element.
    The safe way to specify *IN* directives is to add them to *where_data*, as the specifics of each DB flavor
    will then be properly dealt with.

    The syntax specific to *where_data*'s key/value pairs is as follows:
        1. *key*:
            - an attribute (possibly aliased), or
            - a 2/3-tuple with an attribute, the corresponding SQL comparison operation
              ("=", ">", "<", ">=", "<=", "<>", "in", "like", "ilike", "between" - defaults to "="), and
              a SQL logical operator relating it to the next item ("and", "or" - defaults to "and")
        2. *value*:
            - a scalar, or a list, or an expression possibly containing other attribute(s)

    If not positive integers, *min_count*, *max_count*, *offset_count*, and *limit_count* are ignored.
    If both *min_count* and *max_count* are specified with equal values, then exactly that number of
    tuples must be returned by the query. The parameter *offset_count* is used to offset the retrieval
    of tuples. For both *offset_count* and *limit_count* to be used together with SQLServer, an *ORDER BY*
    clause must have been specifed, otherwise a runtime error is raised.

    If the search is empty, an empty list is returned.
    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param sel_stmt: SELECT command for the query operation
    :param where_clause: optional criteria for tuple selection (ignored if *sel_stmt* contains a *WHERE* clause)
    :param where_vals: values to be associated with the selection criteria
    :param where_data: selection criteria specified as key-value pairs
    :param orderby_clause: optional retrieval order (ignored if *sel_stmt* contains a *ORDER BY* clause)
    :param min_count: optionally defines the minimum number of tuples expected
    :param max_count: optionally defines the maximum number of tuples expected
    :param offset_count: number of tuples to skip (defaults to none)
    :param limit_count: limit to the number of tuples returned, to be specified in the query statement itself
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: list of tuples containing the search result, *[]* on empty search, or *None* if error
    """
    # initialize the return variable
    result: list[tuple] | None = None

    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)
    if engine:
        # process search data provided as key-value pairs
        if where_clause or where_data or orderby_clause:
            sel_stmt, where_vals = _combine_search_data(query_stmt=sel_stmt,
                                                        where_clause=where_clause,
                                                        where_vals=where_vals,
                                                        where_data=where_data,
                                                        orderby_clause=orderby_clause,
                                                        engine=engine)
        # establish the correct bind tags
        if where_vals and DB_BIND_META_TAG in sel_stmt:
            sel_stmt = db_adjust_placeholders(stmt=sel_stmt,
                                              engine=engine)
        # make sure to have a connection
        curr_conn: Any = connection or db_connect(autocommit=engine == DbEngine.SPANNER,
                                                  engine=engine,
                                                  errors=curr_errors)
        if not curr_errors:
            if engine == DbEngine.MYSQL:
                pass
            elif engine == DbEngine.ORACLE:
                from . import oracle_pomes
                result = oracle_pomes.select(sel_stmt=sel_stmt,
                                             where_vals=where_vals,
                                             min_count=min_count,
                                             max_count=max_count,
                                             offset_count=offset_count,
                                             limit_count=limit_count,
                                             connection=curr_conn,
                                             committable=committable if connection else True,
                                             errors=curr_errors)
            elif engine == DbEngine.POSTGRES:
                from . import postgres_pomes
                result = postgres_pomes.select(sel_stmt=sel_stmt,
                                               where_vals=where_vals,
                                               min_count=min_count,
                                               max_count=max_count,
                                               offset_count=offset_count,
                                               limit_count=limit_count,
                                               connection=curr_conn,
                                               committable=committable if connection else True,
                                               errors=curr_errors)
            elif engine == DbEngine.SQLSERVER:
                from . import sqlserver_pomes
                result = sqlserver_pomes.select(sel_stmt=sel_stmt,
                                                where_vals=where_vals,
                                                min_count=min_count,
                                                max_count=max_count,
                                                offset_count=offset_count,
                                                limit_count=limit_count,
                                                connection=curr_conn,
                                                committable=committable if connection else True,
                                                errors=curr_errors)
            elif engine == DbEngine.SPANNER:
                from . import spanner_pomes
                result = spanner_pomes.select(sel_stmt=sel_stmt,
                                              where_vals=where_vals,
                                              min_count=min_count,
                                              max_count=max_count,
                                              offset_count=offset_count,
                                              limit_count=limit_count,
                                              connection=curr_conn,
                                              committable=committable if connection else True,
                                              errors=curr_errors)
        # close the locally acquired connection
        if curr_conn and not connection:
            db_close(connection=curr_conn,
                     engine=engine)

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def db_insert(insert_stmt: str,
              insert_vals: tuple = None,
              insert_data: dict[str, Any] = None,
              return_cols: dict[str, type] = None,
              engine: DbEngine = None,
              connection: Any = None,
              committable: bool = None,
              errors: list[str] = None) -> tuple | int | None:
    """
    Insert a tuple, with values defined in *insert_vals* and *insert_data*, into the database.

    The optional *return_cols* indicate that the values of the columns therein should be returned.
    This is useful to retrieve values from identity columns (that is, columns whose values at insert time
    are handled by the database).

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param insert_stmt: the INSERT command
    :param insert_vals: values to be inserted
    :param insert_data: data to be inserted as key-value pairs
    :param return_cols: optional columns and respective types, whose values are to be returned
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the values of *return_cols*, the number of inserted tuples (0 ou 1), or *None* if error
    """
    # process insert data provided as key-value pairs
    if insert_data:
        insert_stmt, insert_vals = _combine_insert_data(insert_stmt=insert_stmt,
                                                        insert_vals=insert_vals,
                                                        insert_data=insert_data)
    return db_execute(exc_stmt=insert_stmt,
                      bind_vals=insert_vals,
                      return_cols=return_cols,
                      engine=engine,
                      connection=connection,
                      committable=committable,
                      errors=errors)


def db_update(update_stmt: str,
              update_vals: tuple = None,
              update_data: dict[str, Any] = None,
              where_clause: str = None,
              where_vals: tuple = None,
              where_data: dict[str | tuple |
                               tuple[str,
                                     Literal["=", ">", "<", ">=", "<=", "<>",
                                             "in", "like", "ilike", "between"] | None,
                                     Literal["and", "or"] | None], Any] = None,
              return_cols: dict[str, type] = None,
              min_count: int = None,
              max_count: int = None,
              engine: DbEngine = None,
              connection: Any = None,
              committable: bool = None,
              errors: list[str] = None) -> tuple | int | None:
    """
    Update one or more tuples in the database, as defined by the command *update_stmt*.

    Optional selection criteria may be specified in *where_clause* and *where_vals*, or additionally but
    preferably, by key-value pairs in *where_data*. Care should be exercised if *where_clause* contains *IN*
    directives. In PostgreSQL, the list of values for an attribute with the *IN* directive must be contained
    in a specific tuple, and the operation will break for a list of values containing only 1 element.
    The safe way to specify *IN* directives is to add them to *where_data*, as the specifics of each DB flavor
    will then be properly dealt with.

    The syntax specific to *where_data*'s key/value pairs is as follows:
      1. *key*:
          - an attribute (possibly aliased), or
          - a 2/3-tuple with an attribute, the corresponding SQL comparison operation
            ("=", ">", "<", ">=", "<=", "<>", "in", "like", "ilike", "between" - defaults to "="), and
            a SQL logical operator relating it to the next item ("and", "or" - defaults to "and")
      2. *value*:
          - a scalar, or a list, or an expression possibly containing other attribute(s)

    The optional *return_cols* indicate that the values of the columns therein should be returned.
    The target database engine, specified or default,  must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param update_stmt: the UPDATE command
    :param update_vals: values for the update operation
    :param update_data: update data as key-value pairs
    :param where_clause: optional criteria for tuple selection (ignored if *upd_stmt* contains a *WHERE* clause)
    :param where_vals: values to be associated with the selection criteria
    :param where_data: selection criteria as key-value pairs
    :param return_cols: optional columns and respective types, whose values are to be returned
    :param min_count: optionally defines the minimum number of tuples to be updated
    :param max_count: optionally defines the maximum number of tuples to be updated
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the values of *return_cols*, or the number of updated tuples, or *None* if error
    """
    # initialize the return variable
    result:  tuple | int | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    if engine:
        # process update data provided as key-value pairs
        if update_data:
            update_stmt, update_vals = _combine_update_data(update_stmt=update_stmt,
                                                            update_vals=update_vals,
                                                            update_data=update_data)
        # process search data provided as key-value pairs
        if where_clause or where_data:
            update_stmt, where_vals = _combine_search_data(query_stmt=update_stmt,
                                                           where_clause=where_clause,
                                                           where_vals=where_vals,
                                                           where_data=where_data,
                                                           orderby_clause=None,
                                                           engine=engine)
        # combine 'update' and 'where' bind values
        bind_vals: tuple | None = None
        if update_vals and where_vals:
            bind_vals = update_vals + where_vals
        elif update_vals:
            bind_vals = update_vals
        elif where_vals:
            bind_vals = where_vals

        result = db_execute(exc_stmt=update_stmt,
                            bind_vals=bind_vals,
                            return_cols=return_cols,
                            min_count=min_count,
                            max_count=max_count,
                            engine=engine,
                            connection=connection,
                            committable=committable,
                            errors=errors)
    return result


def db_delete(delete_stmt: str,
              where_clause: str = None,
              where_vals: tuple = None,
              where_data: dict[str | tuple |
                               tuple[str,
                                     Literal["=", ">", "<", ">=", "<=", "<>",
                                             "in", "like", "ilike", "between"] | None,
                                     Literal["and", "or"] | None], Any] = None,
              min_count: int = None,
              max_count: int = None,
              engine: DbEngine = None,
              connection: Any = None,
              committable: bool = None,
              errors: list[str] = None) -> int | None:
    """
    Delete one or more tuples in the database, as defined by the *delete_stmt* command.

    Delete criteria may be specified in *where_clause* and *where_vals*, or additionally but preferably,
    by key-value pairs in *where_data*. Care should be exercised if *where_clause* contains *IN*
    directives. In PostgreSQL, the list of values for an attribute with the *IN* directive must be contained
    in a specific tuple, and the operation will break for a list of values containing only 1 element.
    The safe way to specify *IN* directives is to add them to *where_data*, as the specifics of each DB flavor
    will then be properly dealt with.

    The syntax specific to *where_data*'s key/value pairs is as follows:
        1. *key*:
            - an attribute (possibly aliased), or
            - a 2/3-tuple with an attribute and the corresponding SQL comparison operation
              ("=", ">", "<", ">=", "<=", "<>", "in", "like", "ilike", "between" - defaults to "="), followed
              by a SQL logical operator relating it to the next item ("and", "or" - defaults to "and")
        2. *value*:
            - a scalar, or a list, or an expression possibly containing other attribute(s)

    If not positive integers, *min_count*, *max_count*, *offset_count*, and *limit_count* are ignored.
    If both *min_count* and *max_count* are specified with equal values, then exactly that number of
    tuples must be deleted from the database table.

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param delete_stmt: the DELETE command
    :param where_clause: optional criteria for tuple selection (ignored if *delete_stmt* contains a *WHERE* clause)
    :param where_vals: values to be associated with the selection criteria
    :param where_data: selection criteria as key-value pairs
    :param min_count: optionally defines the minimum number of tuples to be deleted
    :param max_count: optionally defines the maximum number of tuples to be deleted
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the number of deleted tuples, or *None* if error
    """
    # process search data provided as key-value pairs
    if where_clause or where_data:
        engine = _assert_engine(engine=engine)
        if engine:
            delete_stmt, where_vals = _combine_search_data(query_stmt=delete_stmt,
                                                           where_clause=where_clause,
                                                           where_vals=where_vals,
                                                           where_data=where_data,
                                                           orderby_clause=None,
                                                           engine=engine)
    return db_execute(exc_stmt=delete_stmt,
                      bind_vals=where_vals,
                      min_count=min_count,
                      max_count=max_count,
                      engine=engine,
                      connection=connection,
                      committable=committable,
                      errors=errors)


def db_bulk_insert(target_table: str,
                   insert_attrs: list[str],
                   insert_vals: list[tuple],
                   engine: DbEngine = None,
                   connection: Any = None,
                   committable: bool = None,
                   identity_column: str = None,
                   errors: list[str] = None) -> int | None:
    """
    Bulk insert rows into *target_table*, with values of *insert_attrs* defined in *insert_vals*.

    Bulk inserts may require non-standard syntax, depending on the database engine being targeted.
    The number of attributes in *insert_attrs* must match the number of bind values in *insert_vals* tuples.
    Specific handling is required for identity columns (i.e., columns whose values are generated directly
    by the database engine - typically, they are also primary keys), and thus they must be identified
    by *identity_column*, and ommited from *insert_stmt*.

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param target_table: the possibly schema-qualified table to insert into
    :param insert_attrs: the list of table attributes to insert values into
    :param insert_vals: the list of values to be inserted
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param identity_column: column whose values are generated by the database
    :param errors: incidental error messages (might be a non-empty list)
    :return: the number of inserted tuples (1 for postgres), or *None* if error
    """
    # initialize the return variable
    result: int | None = None

    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection
    curr_conn: Any = connection or db_connect(engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        if engine == DbEngine.MYSQL:
            pass
        elif engine == DbEngine.POSTGRES:
            from . import postgres_pomes
            insert_stmt: str = f"INSERT INTO {target_table} ({', '.join(insert_attrs)})"
            # pre-insert handling of identity columns
            if identity_column and insert_stmt.find("OVERRIDING SYSTEM VALUE") < 0:
                insert_stmt += " OVERRIDING SYSTEM VALUE"
            insert_stmt += " VALUES %s"
            # obtain template to handle inserts of null values
            template: str = postgres_pomes.build_typified_template(insert_stmt=insert_stmt,
                                                                   nullable_only=True,
                                                                   connection=curr_conn,
                                                                   errors=curr_errors)
            if not curr_errors:
                result = postgres_pomes.bulk_execute(exc_stmt=insert_stmt,
                                                     exc_vals=insert_vals,
                                                     template=template,
                                                     connection=curr_conn,
                                                     committable=False if identity_column else
                                                     (committable if connection else True),
                                                     errors=curr_errors)
                # post-insert handling of identity columns
                if not curr_errors and identity_column:
                    postgres_pomes.identity_post_insert(insert_stmt=insert_stmt,
                                                        connection=curr_conn,
                                                        committable=committable if connection else True,
                                                        identity_column=identity_column,
                                                        errors=curr_errors)
        elif engine in [DbEngine.ORACLE, DbEngine.SQLSERVER]:
            bind_marks: str = _bind_marks(engine=engine,
                                          start=1,
                                          finish=len(insert_attrs) + 1)
            insert_stmt: str = (f"INSERT INTO {target_table} "
                                f"({', '.join(insert_attrs)}) VALUES({bind_marks})")
            if engine == DbEngine.ORACLE:
                from . import oracle_pomes
                result = oracle_pomes.bulk_execute(exc_stmt=insert_stmt,
                                                   exc_vals=insert_vals,
                                                   connection=curr_conn,
                                                   committable=committable if connection else True,
                                                   errors=curr_errors)
            elif engine == DbEngine.SQLSERVER:
                from . import sqlserver_pomes
                # pre-insert handling of identity columns
                if identity_column:
                    sqlserver_pomes.identity_pre_insert(insert_stmt=insert_stmt,
                                                        connection=curr_conn,
                                                        errors=curr_errors)
                if not curr_errors:
                    result = sqlserver_pomes.bulk_execute(exc_stmt=insert_stmt,
                                                          exc_vals=insert_vals,
                                                          connection=curr_conn,
                                                          committable=False if identity_column else
                                                          (committable if connection else True),
                                                          errors=curr_errors)
                    # post-insert handling of identity columns
                    if not curr_errors and identity_column:
                        from . import sqlserver_pomes
                        sqlserver_pomes.identity_post_insert(insert_stmt=insert_stmt,
                                                             connection=curr_conn,
                                                             committable=committable if connection else True,
                                                             identity_column=identity_column,
                                                             errors=curr_errors)
    # close the locally acquired connection
    if curr_conn and not connection:
        db_close(connection=curr_conn,
                 engine=engine)

    # log the operation
    if isinstance(result, int):
        logger: Logger = _DB_LOGGERS[engine]
        if logger:
            logger.debug(f"INSERTed {result} rows into table {engine}.{target_table}")

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def db_bulk_update(target_table: str,
                   set_attrs: list[str],
                   where_attrs: list[str],
                   update_vals: list[tuple],
                   engine: DbEngine = None,
                   connection: Any = None,
                   committable: bool = None,
                   errors: list[str] = None) -> int | None:
    """
    Bulk update rows in *target_table*, with values of *where_attrs* defined in *update_vals*.

    Bulk updates require non-standard syntax, specific for the database engine being targeted.
    The number of attributes in *set_attrs*, plus the number of attributes in *where_attrs*,
    must match the number of bind values in *update_vals* tuples. Note that within *update_vals*,
    the bind values for the *WHERE* clause will follow the ones for the *SET* clause.

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param target_table: the possibly schema-qualified table to update
    :param set_attrs: the list of table attributes to update
    :param where_attrs: the list of table attributes identifying the tuples
    :param update_vals: the list of values to update the database with, and to identify the tuples
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the number of updated tuples, or *None* if error
    """
    # initialize the return variable
    result: int | None = None

    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection
    curr_conn: Any = connection or db_connect(engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        if engine == DbEngine.MYSQL:
            pass
        elif engine == DbEngine.POSTGRES:
            from . import postgres_pomes
            set_items: str = ""
            for set_attr in set_attrs:
                set_items += f"{set_attr} = data.{set_attr}, "
            where_items: str = ""
            for where_attr in where_attrs:
                where_items += f"{target_table}.{where_attr} = data.{where_attr} AND "
            update_stmt: str = (f"UPDATE {target_table}"
                                f" SET {set_items[:-2]} "
                                f"FROM (VALUES %s) AS data ({', '.join(set_attrs + where_attrs)}) "
                                f"WHERE {where_items[:-5]}")
            # modify statement to handle updates of null values
            update_stmt = postgres_pomes.tipify_bulk_update(update_stmt=update_stmt,
                                                            nullable_only=True,
                                                            connection=curr_conn,
                                                            errors=curr_errors)
            if not curr_errors:
                result = postgres_pomes.bulk_execute(exc_stmt=update_stmt,
                                                     exc_vals=update_vals,
                                                     template=None,
                                                     connection=curr_conn,
                                                     committable=committable if connection else True,
                                                     errors=curr_errors)
        elif engine in [DbEngine.ORACLE, DbEngine.SQLSERVER]:
            set_items: str = _bind_columns(engine=engine,
                                           columns=set_attrs,
                                           concat=", ",
                                           start_index=1)
            where_items: str = _bind_columns(engine=engine,
                                             columns=where_attrs,
                                             concat=" AND ",
                                             start_index=len(set_attrs)+1)
            update_stmt: str = f"UPDATE {target_table} SET {set_items} WHERE {where_items}"
            if engine == DbEngine.ORACLE:
                from . import oracle_pomes
                result = oracle_pomes.bulk_execute(exc_stmt=update_stmt,
                                                   exc_vals=update_vals,
                                                   connection=curr_conn,
                                                   committable=committable if connection else True,
                                                   errors=curr_errors)
            else:
                from . import sqlserver_pomes
                result = sqlserver_pomes.bulk_execute(exc_stmt=update_stmt,
                                                      exc_vals=update_vals,
                                                      connection=curr_conn,
                                                      committable=committable if connection else True,
                                                      errors=curr_errors)
    # close the locally acquired connection
    if curr_conn and not connection:
        db_close(connection=curr_conn,
                 engine=engine)

    # log the operation
    if isinstance(result, int):
        logger: Logger = _DB_LOGGERS[engine]
        if logger:
            logger.debug(f"UPDATEd {result} rows on {engine}.{target_table}")

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def db_bulk_delete(target_table: str,
                   where_attrs: list[str],
                   where_vals: list[tuple],
                   engine: DbEngine = None,
                   connection: Any = None,
                   committable: bool = None,
                   errors: list[str] = None) -> int | None:
    """
    Bulk delete from *target_table*, with values of *where_attrs* defined in *where_vals*.

    Bulk deletes may require non-standard syntax, depending on the database engine being targeted.
    The number of attributes in *where_attrs* must match the number of bind values in *where_vals* tuples.

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param target_table: the possibly schema-qualified table to delete from
    :param where_attrs: the list of attributes for identifying the tuples to be deleted
    :param where_vals: the list of values to bind to the attributes
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the number of inserted tuples (1 for postgres), or *None* if error
    """
    # initialize the return variable
    result: int | None = None

    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection
    curr_conn: Any = connection or db_connect(engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        if engine == DbEngine.MYSQL:
            pass
        elif engine == DbEngine.POSTGRES:
            from . import postgres_pomes
            delete_stmt: str = (f"DELETE FROM {target_table} "
                                f"WHERE ({', '.join(where_attrs)}) IN (%s)")
            result = postgres_pomes.bulk_execute(exc_stmt=delete_stmt,
                                                 exc_vals=where_vals,
                                                 template=None,
                                                 connection=curr_conn,
                                                 committable=committable if connection else True,
                                                 errors=curr_errors)
        elif engine in [DbEngine.ORACLE, DbEngine.SQLSERVER]:
            where_items: str = _bind_columns(engine=engine,
                                             columns=where_attrs,
                                             concat=" AND",
                                             start_index=1)
            delete_stmt: str = f"DELETE FROM {target_table} WHERE {where_items}"
            if engine == DbEngine.ORACLE:
                from . import oracle_pomes
                result = oracle_pomes.bulk_execute(exc_stmt=delete_stmt,
                                                   exc_vals=where_vals,
                                                   connection=curr_conn,
                                                   committable=committable if connection else True,
                                                   errors=curr_errors)
            else:
                from . import sqlserver_pomes
                result = sqlserver_pomes.bulk_execute(exc_stmt=delete_stmt,
                                                      exc_vals=where_vals,
                                                      connection=curr_conn,
                                                      committable=committable if connection else True,
                                                      errors=curr_errors)
    # close the locally acquired connection
    if curr_conn and not connection:
        db_close(connection=curr_conn,
                 engine=engine)

    # log the operation
    if isinstance(result, int):
        logger: Logger = _DB_LOGGERS[engine]
        if logger:
            logger.debug(f"DELETEd {result} rows from table {engine}.{target_table}")

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def db_update_lob(lob_table: str,
                  lob_column: str,
                  pk_columns: list[str],
                  pk_vals: tuple,
                  lob_data: bytes | str | Path | BinaryIO,
                  chunk_size: int,
                  engine: DbEngine = None,
                  connection: Any = None,
                  committable: bool = None,
                  errors: list[str] = None) -> None:
    """
    Update a large binary object (LOB) in the given table and column.

    The data for the update may come from *bytes*, from a *Path* or its string representation,
    or from a pointer obtained from *BytesIO* or *Path.open()* in binary mode.

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param lob_table: the table to be update with the new LOB
    :param lob_column: the column to be updated with the new LOB
    :param pk_columns: columns making up a primary key, or a unique identifier for the tuple
    :param pk_vals: values with which to locate the tuple to be updated
    :param lob_data: the LOB data (bytes, a file path, or a file pointer)
    :param chunk_size: size in bytes of the data chunk to read/write, or 0 or *None* for no limit
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: number of LOBs effectively copied, or *None* if error
    """
    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection
    curr_conn: Any = connection or db_connect(engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        if engine == DbEngine.MYSQL:
            pass
        elif engine == DbEngine.ORACLE:
            from . import oracle_pomes
            oracle_pomes.update_lob(lob_table=lob_table,
                                    lob_column=lob_column,
                                    pk_columns=pk_columns,
                                    pk_vals=pk_vals,
                                    lob_data=lob_data,
                                    chunk_size=chunk_size,
                                    connection=curr_conn,
                                    committable=committable if connection else True,
                                    errors=curr_errors)
        elif engine == DbEngine.POSTGRES:
            from . import postgres_pomes
            postgres_pomes.update_lob(lob_table=lob_table,
                                      lob_column=lob_column,
                                      pk_columns=pk_columns,
                                      pk_vals=pk_vals,
                                      lob_data=lob_data,
                                      chunk_size=chunk_size,
                                      connection=curr_conn,
                                      committable=committable if connection else True,
                                      errors=curr_errors)
        elif engine == DbEngine.SQLSERVER:
            from . import sqlserver_pomes
            sqlserver_pomes.update_lob(lob_table=lob_table,
                                       lob_column=lob_column,
                                       pk_columns=pk_columns,
                                       pk_vals=pk_vals,
                                       lob_data=lob_data,
                                       chunk_size=chunk_size,
                                       connection=curr_conn,
                                       committable=committable if connection else True,
                                       errors=curr_errors)
    # close the locally acquired connection
    if curr_conn and not connection:
        db_close(connection=curr_conn,
                 engine=engine)

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)


def db_execute(exc_stmt: str,
               bind_vals: tuple = None,
               return_cols: dict[str, type] = None,
               min_count: int = None,
               max_count: int = None,
               engine: DbEngine = None,
               connection: Any = None,
               committable: bool = None,
               errors: list[str] = None) -> tuple | int | None:
    """
    Execute the command *exc_stmt* on the database.

    This command might be a DML ccommand modifying the database, such as inserting, updating or
    deleting tuples, or it might be a DDL statement, or it might even be an environment-related command.

    The optional bind values for this operation are in *bind_vals*. The optional *return_cols* indicate that
    the values of the columns therein should be returned upon execution of *exc_stmt*. This is typical for
    *INSERT* or *UPDATE* statements on tables with *identity-type* columns, which are columns whose values
    are generated by the database itself. Otherwise, the value returned is the number of inserted, modified,
    or deleted tuples, or *None* if an error occurred.

    The value returned by this operation (as *cursor.rowcount*) is verified against *min_count* or *max_count*,
    if provided. An error is issued if a disagreement exists, followed by a rollback. This is an optional feature,
    intended to minimize data loss due to programming mistakes.

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param exc_stmt: the command to execute
    :param bind_vals: optional bind values
    :param return_cols: optional columns and respective types, whose values are to be returned
    :param min_count: optionally defines the minimum number of tuples to be affected
    :param max_count: optionally defines the maximum number of tuples to be affected
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the values of *return_cols*, the return value from the command execution, or *None* if error
    """
    # initialize the return variable
    result: tuple | int | None = None

    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection
    curr_conn: Any = connection or db_connect(autocommit=engine == DbEngine.SPANNER,
                                              engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        # establish the correct bind tags
        if bind_vals and DB_BIND_META_TAG in exc_stmt:
            exc_stmt = db_adjust_placeholders(stmt=exc_stmt,
                                              engine=engine)
        if engine == DbEngine.MYSQL:
            pass
        elif engine == DbEngine.ORACLE:
            from . import oracle_pomes
            result = oracle_pomes.execute(exc_stmt=exc_stmt,
                                          bind_vals=bind_vals,
                                          return_cols=return_cols,
                                          min_count=min_count,
                                          max_count=max_count,
                                          connection=curr_conn,
                                          committable=committable if connection else True,
                                          errors=curr_errors)
        elif engine == DbEngine.POSTGRES:
            from . import postgres_pomes
            result = postgres_pomes.execute(exc_stmt=exc_stmt,
                                            bind_vals=bind_vals,
                                            return_cols=return_cols,
                                            min_count=min_count,
                                            max_count=max_count,
                                            connection=curr_conn,
                                            committable=committable if connection else True,
                                            errors=curr_errors)
        elif engine == DbEngine.SQLSERVER:
            from . import sqlserver_pomes
            result = sqlserver_pomes.execute(exc_stmt=exc_stmt,
                                             bind_vals=bind_vals,
                                             return_cols=return_cols,
                                             min_count=min_count,
                                             max_count=max_count,
                                             connection=curr_conn,
                                             committable=committable if connection else True,
                                             errors=curr_errors)
        elif engine == DbEngine.SPANNER:
            from . import spanner_pomes
            result = spanner_pomes.execute(exc_stmt=exc_stmt,
                                           bind_vals=bind_vals,
                                           return_cols=return_cols,
                                           min_count=min_count,
                                           max_count=max_count,
                                           connection=curr_conn,
                                           committable=committable if connection else True,
                                           errors=curr_errors)
    # close the locally acquired connection
    if curr_conn and not connection:
        db_close(connection=curr_conn,
                 engine=engine)

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def db_call_function(func_name: str,
                     func_vals: tuple = None,
                     engine: DbEngine = None,
                     connection: Any = None,
                     committable: bool = None,
                     errors: list[str] = None) -> list[tuple] | None:
    """
    Execute the stored function *func_name* in the database, with the parameters given in *func_vals*.

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param func_name: name of the stored function
    :param func_vals: parameters for the stored function
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the data returned by the function, or *None* if error
    """
    # initialize the return variable
    result: Any = None

    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection
    curr_conn: Any = connection or db_connect(engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        if engine == DbEngine.MYSQL:
            pass
        elif engine == DbEngine.ORACLE:
            from . import oracle_pomes
            result = oracle_pomes.call_function(func_name=func_name,
                                                func_vals=func_vals,
                                                connection=curr_conn,
                                                committable=committable if connection else True,
                                                errors=curr_errors)
        elif engine == DbEngine.POSTGRES:
            from . import postgres_pomes
            result = postgres_pomes.call_procedure(proc_name=func_name,
                                                   proc_vals=func_vals,
                                                   connection=curr_conn,
                                                   committable=committable if connection else True,
                                                   errors=curr_errors)
        elif engine == DbEngine.SQLSERVER:
            from . import sqlserver_pomes
            result = sqlserver_pomes.call_procedure(proc_name=func_name,
                                                    proc_vals=func_vals,
                                                    connection=curr_conn,
                                                    committable=committable if connection else True,
                                                    errors=curr_errors)
    # close the locally acquired connection
    if curr_conn and not connection:
        db_close(connection=curr_conn,
                 engine=engine)

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def db_call_procedure(proc_name: str,
                      proc_vals: tuple = None,
                      engine: DbEngine = None,
                      connection: Any = None,
                      committable: bool = None,
                      errors: list[str] = None) -> list[tuple] | None:
    """
    Execute the stored procedure *proc_name* in the database, with the parameters given in *proc_vals*.

    The target database engine, specified or default, must have been previously configured.
    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param proc_name: name of the stored procedure
    :param proc_vals: parameters for the stored procedure
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the data returned by the procedure, or *None* if error
    """
    # initialize the return variable
    result: Any = None

    # necessary, lest 'errors' be passed to function requiring it to be empty or null
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection
    curr_conn: Any = connection or db_connect(engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        if engine == DbEngine.MYSQL:
            pass
        elif engine == DbEngine.ORACLE:
            from . import oracle_pomes
            result = oracle_pomes.call_procedure(proc_name=proc_name,
                                                 proc_vals=proc_vals,
                                                 connection=curr_conn,
                                                 committable=committable if connection else True,
                                                 errors=curr_errors)
        elif engine == DbEngine.POSTGRES:
            from . import postgres_pomes
            result = postgres_pomes.call_procedure(proc_name=proc_name,
                                                   proc_vals=proc_vals,
                                                   connection=curr_conn,
                                                   committable=committable if connection else True,
                                                   errors=curr_errors)
        elif engine == DbEngine.SQLSERVER:
            from . import sqlserver_pomes
            result = sqlserver_pomes.call_procedure(proc_name=proc_name,
                                                    proc_vals=proc_vals,
                                                    connection=curr_conn,
                                                    committable=committable if connection else True,
                                                    errors=curr_errors)
    # close the locally acquired connection
    if curr_conn and not connection:
        db_close(connection=curr_conn,
                 engine=engine)

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def __get_version(engine: DbEngine,
                  connection: Any) -> str:
    """
    Obtain and return the current version of *engine*.

    :param engine: the reference database engine (the default engine, if not provided)
    :return: the engine's current version, or an empty string if not found
    """
    # initialize the return variable
    result: str = ""

    stmt: str | None = None
    match engine:
        case DbEngine.MYSQL:
            pass
        case DbEngine.ORACLE:
            stmt = "SELECT version FROM v$instance"
        case DbEngine.POSTGRES:
            stmt = "SHOW server_version"
        case DbEngine.SQLSERVER:
            stmt = "SELECT @@VERSION"

    if stmt:
        reply: list[tuple[str]] = db_select(sel_stmt=stmt,
                                            engine=engine,
                                            connection=connection)
        if reply:
            result = reply[0][0]

    return result



import time
from collections.abc import Generator
from contextlib import suppress
from pypomes_core import timestamp_duration
from typing import Any

from .db_common import _DB_LOGGERS, DbEngine, _assert_engine, _except_msg
from .db_pomes import db_connect, db_close


def db_stream_data(table: str,
                   columns: list[str],
                   where_clause: str = None,
                   orderby_clause: str = None,
                   offset_count: int = None,
                   limit_count: int = None,
                   batch_size_in: int = None,
                   batch_size_out: int = None,
                   engine: DbEngine = None,
                   connection: Any = None,
                   committable: bool = None,
                   errors: list[str] = None) -> Generator[list[tuple], None, None]:
    """
    Stream data from a database table.

    This is accomplished with the implementation of the *generator* pattern, whereby an *iterator*
    is returned, allowing the invoker to iterate over the values being streamed.
    The database in *engine* must be in the list of databases configured and supported by this package.

    It is recommend that *orderby_clause* be provided, specially if partial streaming is being requested.
    If not positive integers, *offset_count*, *limit_count*, *batch_size_in*, and *chunk_size_out* are ignored.
    Care should be exercised when specifying *offset_count*, so as not to skip wanted tuples,
    as it is used to offset the retrieval of tuples.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param table: the, possibly schema-qualified, table to read the data from
    :param columns: the table columns to stream
    :param where_clause: the criteria for tuple selection
    :param orderby_clause: optional retrieval order
    :param offset_count: number of tuples to skip in source table (defaults to none)
    :param limit_count: maximum number of tuples to migrate (defaults to no limit)
    :param batch_size_in: maximum number of tuples to read in each batch (defaults to no limit)
    :param batch_size_out: maximum number of tuples to stream in each batch (defaults to no limit)
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit upon errorless completion
    :param errors: incidental error messages (might be a non-empty list)
    """
    # initialize a local errors list
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection to the source database
    curr_conn: Any = connection or db_connect(engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        # normalize these parameters
        offset_count = __normalize_value(value=offset_count)
        limit_count = __normalize_value(value=limit_count)
        batch_size_in = __normalize_value(value=batch_size_in)
        batch_size_out = __normalize_value(value=batch_size_out)
        if 0 < limit_count < batch_size_in:
            batch_size_in = limit_count
            if _DB_LOGGERS[engine]:
                _DB_LOGGERS[engine].debug(msg=f"Value of input batch size changed to {batch_size_in}")
        if 0 < batch_size_in < batch_size_out:
            batch_size_out = batch_size_in
            if _DB_LOGGERS[engine]:
                _DB_LOGGERS[engine].debug(msg=f"Value of output batch size changed to {batch_size_out}")

        # buid the SELECT query
        sel_stmt: str = f"SELECT {', '.join(columns)} FROM {table}"
        if where_clause:
            sel_stmt += f" WHERE {where_clause}"
        if orderby_clause:
            sel_stmt += f" ORDER BY {orderby_clause}"
        elif _DB_LOGGERS[engine] and (offset_count or batch_size_in):
            _DB_LOGGERS[engine].warning(msg="Attempting partial data migration without "
                                            "providing an ORDER BY clause is not recommended")
        sel_stmt += " OFFSET @offset ROWS"

        if limit_count or batch_size_in:
            if engine == DbEngine.POSTGRES:
                sel_stmt += " LIMIT @limit"
            elif engine in [DbEngine.ORACLE, DbEngine.SQLSERVER]:
                sel_stmt += " FETCH @limit ROWS ONLY"

        # log the streaming start
        if _DB_LOGGERS[engine]:
            _DB_LOGGERS[engine].debug(msg=f"Started streaming data from {engine}.{table}, limit "
                                          f"{limit_count}, offset {offset_count}, batch size in {batch_size_in}, "
                                          f"batch size out {batch_size_out}, connection {id(curr_conn)}")
        # stream the tuples
        row_count: int = 0
        stream_count: int = 0  # number of tuples effectively streamed
        err_msg: str | None = None
        try:
            source_cursor: Any = curr_conn.cursor()

            # adjust the offset and limit
            curr_stmt: str = sel_stmt
            curr_limit: int = min(batch_size_in, limit_count)
            if curr_limit == 0:
                curr_limit = max(batch_size_in, limit_count)
            if curr_limit:
                curr_stmt = curr_stmt.replace("@limit", str(curr_limit), 1)
            else:
                curr_stmt = curr_stmt.replace(" LIMIT @limit", "", 1)
                curr_stmt = curr_stmt.replace(" FETCH @limit ROWS ONLY", "", 1)
            if offset_count:
                curr_stmt = curr_stmt.replace("@offset", str(offset_count), 1)\
                                     .replace(" FETCH ", " FETCH NEXT ", 1)
            else:
                curr_stmt = curr_stmt.replace(" OFFSET @offset ROWS", "", 1)\
                                     .replace(" FETCH ", " FETCH FIRST ", 1)
            source_cursor.execute(curr_stmt)
            rows_in: list[tuple] = source_cursor.fetchall()

            # traverse the result set
            while rows_in:
                # log the retrieval operation
                if _DB_LOGGERS[engine]:
                    _DB_LOGGERS[engine].debug(msg=f"Read {len(rows_in)} tuples from {engine}.{table}, "
                                                  f"offset {offset_count + stream_count}, connection {id(curr_conn)}")
                pos_from: int = 0

                # migrate the tuples
                while pos_from < len(rows_in):
                    pos_to: int = min(pos_from + batch_size_out, len(rows_in)) \
                                  if batch_size_out else len(rows_in)
                    rows_out: list[tuple] = rows_in[pos_from:pos_to]
                    yield rows_out
                    pos_from = pos_to

                    # increment the tuple streaming counter and log the partial streaming
                    stream_count += len(rows_out)
                    if _DB_LOGGERS[engine]:
                        _DB_LOGGERS[engine].debug(msg=f"Streamed {stream_count} tuples "
                                                      f"from {engine}.{table}, connection {id(curr_conn)}")
                # read the next batch
                if limit_count > stream_count or (batch_size_in and not limit_count):
                    curr_limit = min(batch_size_in, limit_count - stream_count)
                    if curr_limit <= 0:
                        curr_limit = max(batch_size_in, limit_count - stream_count)
                    curr_stmt = sel_stmt.replace("@offset", str(offset_count + stream_count), 1)\
                                        .replace("@limit", str(curr_limit), 1)\
                                        .replace(" FETCH ", " FETCH NEXT ", 1)
                    source_cursor.execute(statement=curr_stmt)
                    rows_in = source_cursor.fetchall()
                else:
                    # signal end of migration
                    rows_in = []

            # close the cursors and commit the transactions
            source_cursor.close()
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            # rollback the transactions
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            row_count = 0
            err_msg = _except_msg(exception=e,
                                  connection=curr_conn,
                                  engine=engine)
        finally:
            # close the connections, if locally acquired
            if curr_conn and not connection:
                db_close(connection=curr_conn,
                         engine=engine)

        # log the streaming finish
        if err_msg:
            curr_errors.append(err_msg)
            if _DB_LOGGERS[engine]:
                _DB_LOGGERS[engine].error(msg=err_msg)
        elif not curr_errors and _DB_LOGGERS[engine]:
            _DB_LOGGERS[engine].debug(msg=f"Finished streaming {row_count} tuples "
                                          f"from {engine}.{table}, connection {id(curr_conn)}")

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)


def db_stream_lobs(table: str,
                   lob_column: str,
                   pk_columns: list[str] = None,
                   ret_column: str = None,
                   where_clause: str = None,
                   orderby_clause: str = None,
                   offset_count: int = None,
                   limit_count: int = None,
                   batch_size: int = None,
                   chunk_size: int = None,
                   engine: DbEngine = None,
                   connection: Any = None,
                   committable: bool = None,
                   errors: list[str] = None,
                   log_trigger: int = 10000) -> Generator[dict[str, Any] | str | bytes | None, None, None]:
    """
    Stream data in large binary objects (LOBs) from a database table.

    This is accomplished with the implementation of the *generator* pattern, whereby an *iterator*
    is returned, allowing the invoker to iterate over the values being streamed.
    The database in *engine* must be in the list of databases configured and supported by this package.
    Either *orderby_clause* or one or more columns making up a primary key, or a unique row identifier,
    must be provided in *pk_columns*. The content of these columns, along with the contents of the
    optional *ret_column*, are returned as metadata on the first data chunk of each tuple.

    If not positive integers, *offset_count*, *limit_count*, *batch_size*, and *chunk_size* are ignored.
    Further, if *batch_size* and *limit_count* are both defined, and *batch_size* is set to a larger value,
    it is reduced accordingly.

    Care should be exercised when specifying *offset_count*, so as not to skip wanted tuples, as it is used
    to offset the retrieval of tuples. Finally, if *batch_size* or *offset_count* has been specified,
    but *orderby_clause* has not, then an ORDER BY clause is constructed from the data in *pk_columns*.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param table: the table holding the LOBs
    :param lob_column: the column holding the LOB
    :param pk_columns: columns making up a primary key, or a unique identifier for a tuple, in database
    :param ret_column: optional column whose content to return as metadata when yielding
    :param where_clause: the criteria for tuple selection
    :param orderby_clause: optional retrieval order
    :param offset_count: number of tuples to skip in source table (defaults to none)
    :param limit_count: maximum number of tuples to migrate (defaults to no limit)
    :param batch_size: maximum number of tuples to read in each batch (defaults to no limit)
    :param chunk_size: size in bytes of the data chunk to read/write (defaults to no limit)
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit upon errorless completion
    :param errors: incidental error messages
    :param log_trigger: number of tuples to trigger logging info on migration (defaults to 10000 tuples)
    """
    # initialize a local errors list
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)

    # make sure to have a connection to the source database
    curr_conn: Any = connection or db_connect(engine=engine,
                                              errors=curr_errors)
    if not curr_errors:
        # normalize these parameters
        offset_count = __normalize_value(value=offset_count)
        limit_count = __normalize_value(value=limit_count)
        batch_size = __normalize_value(value=batch_size)
        chunk_size = __normalize_value(value=chunk_size) or -1
        log_trigger = __normalize_value(value=log_trigger)
        if 0 < limit_count < batch_size:
            batch_size = limit_count
            if _DB_LOGGERS[engine]:
                _DB_LOGGERS[engine].debug(msg=f"Value of batch size changed to {batch_size}")

        # buid the SELECT query
        ref_columns: list[str] = pk_columns.copy() if pk_columns else []
        lob_index: int = len(pk_columns)
        sel_stmt: str = f"SELECT {', '.join(pk_columns)}"
        if ret_column and ret_column not in ref_columns:
            sel_stmt += f", {ret_column}"
            lob_index += 1
            ref_columns.append(ret_column)
        sel_stmt += f", {lob_column} FROM {table}"
        if where_clause:
            sel_stmt += f" WHERE {where_clause}"
        if not orderby_clause and pk_columns and (offset_count or batch_size):
            orderby_clause = ", ".join(pk_columns)
        if orderby_clause:
            sel_stmt += f" ORDER BY {orderby_clause}"
        elif _DB_LOGGERS[engine] and (offset_count or batch_size):
            _DB_LOGGERS[engine].warning(msg="Attempting partial LOB migration without providing "
                                            "PK columns or an ORDER BY clause is not recommended")
        sel_stmt += " OFFSET @offset ROWS"

        if limit_count or batch_size:
            if engine == DbEngine.POSTGRES:
                sel_stmt += " LIMIT @limit"
            elif engine in [DbEngine.ORACLE, DbEngine.SQLSERVER]:
                sel_stmt += " FETCH @limit ROWS ONLY"

        # log the migration start
        if _DB_LOGGERS[engine]:
            _DB_LOGGERS[engine].debug(msg="Started streaming LOBs "
                                          f"from {engine}.{table}.{lob_column}, limit {limit_count}, "
                                          f"offset {offset_count}, batch size {batch_size}, "
                                          f"chunk size {chunk_size}, connection {id(curr_conn)}")
        # stream the LOBs
        byte_count: int = 0
        byte_step: int = 0
        row_count: int = 0
        row_step: int = 0
        start_count: float = time.time()
        start_step: float = time.time()
        err_msg: str | None = None
        try:
            source_cursor: Any = curr_conn.cursor()

            # adjust the offset and limit
            curr_stmt: str = sel_stmt
            curr_limit: int = min(batch_size, limit_count)
            if curr_limit == 0:
                curr_limit = max(batch_size, limit_count)
            if curr_limit:
                curr_stmt = curr_stmt.replace("@limit", str(curr_limit), 1)
            else:
                curr_stmt = curr_stmt.replace(" LIMIT @limit", "", 1)
                curr_stmt = curr_stmt.replace(" FETCH @limit ROWS ONLY", "", 1)
            if offset_count:
                curr_stmt = (curr_stmt.replace("@offset", str(offset_count), 1)
                                      .replace(" FETCH ", " FETCH NEXT ", 1))
            else:
                curr_stmt = (curr_stmt.replace(" OFFSET @offset ROWS", "", 1)
                                      .replace(" FETCH ", " FETCH FIRST ", 1))
            # go for the data
            next_rs: bool = True
            while next_rs:
                next_rs = False

                # execute the query
                # (parameter name is 'statement' for Oracle, 'query' for PostgreSQL, 'sql' for SQLServer)
                source_cursor.execute(curr_stmt)

                # traverse the result set
                row: tuple = source_cursor.fetchone()
                while row:
                    next_rs = True

                    # retrieve the values of the primary key and reference columns (leave LOB column out)
                    ref_vals: tuple = tuple([row[inx] for inx in range(lob_index)])
                    identifier: dict[str, Any] = {}
                    for inx, pk_val in enumerate(ref_vals):
                        identifier[ref_columns[inx]] = pk_val
                    # send the LOB's metadata
                    yield identifier

                    # access the LOB's bytes in chunks and stream them
                    first: bool = True
                    offset: int = 1
                    is_migrated: bool = False
                    lob: Any = row[lob_index]
                    lob_data: bytes | str = lob.read(offset=offset,
                                                     amount=chunk_size) if lob is not None else None
                    # make sure to skip null LOBs, and to migrate empty ones
                    while lob_data is not None and (first or len(lob_data) > 0):
                        first = False
                        is_migrated = True
                        # send a data chunk
                        yield lob_data

                        # read the next chunk
                        if len(lob_data) > 0:
                            offset += len(lob_data)
                            lob_data = lob.read(offset=offset,
                                                amount=chunk_size)
                    if is_migrated:
                        # increment the LOB migration counter, if applicable
                        byte_count += offset - 1
                        byte_step += offset - 1
                        row_count += 1
                        row_step += 1

                    # signal that sending data chunks for the current LOB is finished
                    yield None

                    # log partial result at each 'log_trigger' LOBs migrated
                    if _DB_LOGGERS[engine] and row_step >= log_trigger:
                        finish_step: float = time.time()
                        mins: float = (finish_step - start_step) / 60
                        duration: str = timestamp_duration(start=start_step,
                                                           finish=finish_step)
                        _DB_LOGGERS[engine].debug(msg=f"Streamed {row_step} LOBs, {byte_step} bytes, in "
                                                      f"{duration} ({row_step/mins:.2f} LOBs/min, "
                                                      f"{byte_step/(mins * 1024 ** 2):.2f} MBytes/min),  "
                                                      f"from {engine}.{table}.{lob_column}, , connection "
                                                      f"{id(curr_conn)} offset {offset_count + row_count - row_step}")
                        byte_step = 0
                        row_step = 0
                        start_step = finish_step

                    # retrieve the next row
                    row = source_cursor.fetchone()

                # adjust the new offset and limit
                if next_rs:
                    if limit_count > row_count or (batch_size and not limit_count):
                        curr_limit = min(batch_size, limit_count - row_count)
                        if curr_limit <= 0:
                            curr_limit = max(batch_size, limit_count - row_count)
                        curr_stmt = sel_stmt.replace("@offset", str(offset_count + row_count), 1)\
                                            .replace("@limit", str(curr_limit), 1)\
                                            .replace(" FETCH ", " FETCH NEXT ", 1)
                    else:
                        # signal end of migration
                        next_rs = False

            # close the cursors and commit the transactions
            source_cursor.close()
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            # rollback the transactions
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            err_msg = _except_msg(exception=e,
                                  connection=curr_conn,
                                  engine=engine)
        finally:
            # close the connections, if locally acquired
            if curr_conn and not connection:
                db_close(connection=curr_conn,
                         engine=engine)
        # log the finish
        if err_msg:
            curr_errors.append(err_msg)
            if _DB_LOGGERS[engine]:
                _DB_LOGGERS[engine].error(msg=err_msg)
        elif not curr_errors and _DB_LOGGERS[engine]:
            finish_count: float = time.time()
            mins = (finish_count - start_count) / 60
            duration: str = timestamp_duration(start=start_count,
                                               finish=finish_count)
            _DB_LOGGERS[engine].debug(msg=f"Finished streaming {row_count} LOBs, {byte_count} bytes, "
                                          f"in {duration} ({row_count/mins:.2f} LOBs/min, "
                                          f"{byte_count/(mins * 1024 ** 2):.2f} MBytes/min), "
                                          f"from {engine}.{table}.{lob_column}, "
                                          f"offset {offset_count}, connection {id(curr_conn)}")

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)


def __normalize_value(value: int) -> int:
    """
    Normalize *value* to a consistent value.

    :param value: the value to normalized
    :return: the normalized value
    """
    if isinstance(value, int) and \
            not isinstance(value, bool) and value > 0:
        result: int = value
    else:
        result: int = 0

    return result



from typing import Any

from .db_common import (
    _DB_CONN_DATA, DbEngine, DbParam, _assert_engine, _get_param
)
from .db_pomes import db_execute, db_select


def db_create_session_table(engine: DbEngine,
                            connection: Any,
                            table_name: str,
                            column_data: list[str],
                            errors: list[str] = None) -> None:
    """
    Create the session-scoped table *table_name*, with the list of columns specifications in *column_data*.

    The table created has the scope of the given *connection*, its data is kept throughout the session,
    and is automatically dropped when the session ends, that is, when *connection* is closed.

    Note that there is no support for session-scoped temporary tables for versions of *Oracle* before 18c,
    and the newer versions which do support them require their names to start with *ORA$PTT_*.
    In case these two conditions are not met, a global temporary table is created, instead.
    Although its data are session-specific, the table definition for such table remains in the database
    permanently, and thus the non-existence of a table named *table_name* is verified beforehand.

    For *SQLServer*, session-scoped temporary table names must start with the hash symbol *#*, an error
    being raised if otherwise.

    :param engine: the reference database engine
    :param connection: the connection defining the session
    :param table_name: the, possibly schema-qualified, name of the table to be created
    :param column_data: this list of column information for the table cration
    :param errors: incidental error messages (might be a non-empty list)
    """
    # necessary, lest the state of 'errors' be tested
    curr_errors: list[str] = []

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=curr_errors)
    if engine:
        stmt: str | None = None
        match engine:
            case DbEngine.POSTGRES:
                # - 'TEMP' or 'TEMPORARY' creates a session-scoped table
                # - 'ON COMMIT PRESERVE ROWS' ensures rows are kept until session ends
                stmt = (f"CREATE TEMP TABLE {table_name} "
                        f"({', '.join(column_data)}) ON COMMIT PRESERVE ROWS")
            case DbEngine.MYSQL:
                # 'TEMPORARY' ensures the table is dropped when the session ends
                stmt = f"CREATE TEMPORARY TABLE {table_name} ({', '.join(column_data)})"
            case DbEngine.ORACLE:
                # private, session-scoped, temporary tables require:
                #   - database version 18c or later
                #   - name starting with 'ORA$PTT_'
                version: str = _get_param(engine=engine,
                                          param=DbParam.VERSION)
                if version and version > "18" and table_name.upper().startswith("ORA$PTT_"):
                    # 'ON COMMIT PRESERVE DEFINITION' ensures the table lasts for the session
                    stmt = (f"CREATE PRIVATE TEMPORARY TABLE {table_name.upper()} "
                            f"({', '.join(column_data)}) ON COMMIT PRESERVE DEFINITION")
                elif not db_table_exists(table_name=table_name,
                                         engine=DbEngine.ORACLE,
                                         errors=curr_errors) and not curr_errors:
                    # - the table definition remains in the database permanently, but the data is session-specific
                    # - 'ON COMMIT PRESERVE ROWS' keeps data for the entire session
                    stmt = (f"CREATE GLOBAL TEMPORARY TABLE {table_name} "
                            f"({'. '.join(column_data)}) ON COMMIT PRESERVE ROWS")
            case DbEngine.SQLSERVER:
                # 'table_name' must be prepended with '#', which creates a local, session-scoped, temporary table
                stmt = f"CREATE TABLE {table_name} ({', '.join(column_data)})"

        if stmt:
            db_execute(exc_stmt=stmt,
                       engine=engine,
                       connection=connection,
                       errors=curr_errors)

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)


def db_get_session_table_prefix(engine: DbEngine) -> str:
    """
    Return the prefix required in the table's name for it to be created as a session-scoped table.

    The prefix is returned in uppercase. For contexts which do not require such a prefix, such as
    *MySQL* or *PostgreSQL* engines, or *Oracle* engines prior to version 18c, an empty string is returned.

    :param engine: the reference database engine
    :return: th prefix required for session tables, or an empty string if otherwise
    """
    # initialize the return variable
    result: str = ""

    # assert the database engine
    engine = next(iter(_DB_CONN_DATA)) if not engine and _DB_CONN_DATA else engine

    if engine == DbEngine.ORACLE:
        version: str = _get_param(engine=engine,
                                  param=DbParam.VERSION)
        if version and version > "18":
            result = "ORA$PTT_"
    elif engine == DbEngine.SQLSERVER:
        result = "#"

    return result


def db_get_tables(schema: str = None,
                  engine: DbEngine = None,
                  connection: Any = None,
                  committable: bool = None,
                  errors: list[str] = None) -> list[str] | None:
    """
    Retrieve the list of schema-qualified tables in the database.

    The parameter *committable* is relevant only if *connection* is provided, and is otherwise ignored.
    A rollback is always attempted, if an error occurs.

    :param schema: optional name of the schema to restrict the search to
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit operation on errorless completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the schema-qualified table names found, or *None* if error
    """
    # initialize the return variable
    result: list[str] | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    if engine:
        # build the query
        if engine == DbEngine.ORACLE:
            sel_stmt: str = "SELECT schema_name || '.' || table_name FROM all_tables"
            if schema:
                sel_stmt += f" WHERE owner = '{schema.upper()}'"
        else:
            sel_stmt: str = ("SELECT table_schema || '.' || table_name "
                             "FROM information_schema.tables "
                             "WHERE table_type = 'BASE TABLE'")
            if schema:
                sel_stmt += f" AND LOWER(table_schema) = '{schema.lower()}'"

        # execute the query
        recs: list[tuple[str]] = db_select(sel_stmt=sel_stmt,
                                           engine=engine,
                                           connection=connection,
                                           committable=committable,
                                           errors=errors)
        # process the query result
        if isinstance(recs, list):
            result = [rec[0] for rec in recs]

    return result


def db_table_exists(table_name: str,
                    engine: DbEngine = None,
                    connection: Any = None,
                    committable: bool = None,
                    errors: list[str] = None) -> bool | None:
    """
    Determine whether the table *table_name* exists in the database.

    If *table_name* is schema-qualified, then the search will be restricted to that schema.
    The parameter *committable* is relevant only if *connection* is provided, and is otherwise ignored.
    A rollback is always attempted, if an error occurs.

    :param table_name: the, possibly schema-qualified, name of the table to look for
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit operation on errorless completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: *True* if the table was found, *False* otherwise, or *None* if error
    """
    # initialize the return variable
    result: bool | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    if engine:
        # determine the table schema
        table_schema: str
        table_schema, table_name = table_name.split(sep=".") if "." in table_name else (None, table_name)

        # build the query
        if engine == DbEngine.ORACLE:
            sel_stmt: str = ("SELECT COUNT(*) FROM all_tables "
                             f"WHERE table_name = '{table_name.upper()}'")
            if table_schema:
                sel_stmt += f" AND owner = '{table_schema.upper()}'"
        else:
            sel_stmt: str = ("SELECT COUNT(*) "
                             "FROM information_schema.tables "
                             f"WHERE table_type = 'BASE TABLE' AND "
                             f"LOWER(table_name) = '{table_name.lower()}'")
            if table_schema:
                sel_stmt += f" AND LOWER(table_schema) = '{table_schema.lower()}'"

        # execute the query
        recs: list[tuple[int]] = db_select(sel_stmt=sel_stmt,
                                           engine=engine,
                                           connection=connection,
                                           committable=committable,
                                           errors=errors)
        # process the query result
        if isinstance(recs, list):
            result = recs[0][0] > 0

    return result


def db_drop_table(table_name: str,
                  engine: DbEngine = None,
                  connection: Any = None,
                  committable: bool = None,
                  errors: list[str] = None) -> None:
    """
    Drop the table given by the, possibly schema-qualified, *table_name*.

    This is a silent *DDL* operation. Whether commits or rollbacks are applicable,
    and what their use would entail, depends on the response of the *engine* to the
    mixing of *DDL* and *DML* statements in a transaction.

    The parameter *committable* is relevant only if *connection* is provided, and is otherwise ignored.
    A rollback is always attempted, if an error occurs.

    :param table_name: the, possibly schema-qualified, name of the table to drop
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit operation on errorless completion
    :param errors: incidental error messages (might be a non-empty list)
    """
    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    if engine:
        # build the DROP statement
        if engine == DbEngine.ORACLE:
            # oracle has no 'IF EXISTS' clause
            drop_stmt: str = \
                (f"BEGIN"
                 f" EXECUTE IMMEDIATE 'DROP TABLE {table_name} CASCADE CONSTRAINTS'; "
                 "EXCEPTION"
                 " WHEN OTHERS THEN NULL; "
                 "END;")
        elif engine == DbEngine.POSTGRES:
            drop_stmt: str = \
                ("DO $$"
                 "BEGIN"
                 f" EXECUTE 'DROP TABLE IF EXISTS {table_name} CASCADE'; "
                 "EXCEPTION"
                 " WHEN OTHERS THEN NULL; "
                 "END $$;")
        elif engine == DbEngine.SQLSERVER:
            drop_stmt: str = \
                ("BEGIN TRY"
                 f" EXEC('DROP TABLE IF EXISTS {table_name} CASCADE;'); "
                 "END TRY "
                 "BEGIN CATCH "
                 "END CATCH;")
        else:
            drop_stmt: str = f"DROP TABLE IF EXISTS {table_name}"

        # drop the table
        db_execute(exc_stmt=drop_stmt,
                   engine=engine,
                   connection=connection,
                   committable=committable,
                   errors=errors)


def db_get_table_ddl(table_name: str,
                     engine: DbEngine = None,
                     connection: Any = None,
                     committable: bool = None,
                     errors: list[str] = None) -> str | None:
    """
    Retrieve the DDL script used to create the table *table_name*.

    Note that *table_name* must be schema-qualified, or else the invocation will fail.
    For *postgres* databases, make sure that the function *pg_get_tabledef* is installed and accessible.
    This function is freely available at https://github.com/MichaelDBA/pg_get_tabledef.

    The parameter *committable* is relevant only if *connection* is provided, and is otherwise ignored.
    A rollback is always attempted, if an error occurs.

    :param table_name: the schema-qualified name of the table
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit operation on errorless completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the DDL script used to create the index, or *None* if error or the table does not exist
    """
    # initialize the return variable
    result: str | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    if engine:
        # determine the table schema
        table_schema: str
        table_schema, table_name = table_name.split(sep=".") if "." in table_name else (None, table_name)

        if table_schema:
            # build the query
            sel_stmt: str | None = None
            if engine == DbEngine.MYSQL:
                pass
            if engine == DbEngine.ORACLE:
                sel_stmt = ("SELECT DBMS_METADATA.GET_DDL('TABLE', "
                            f"'{table_name.upper()}', '{table_schema.upper()}') "
                            "FROM DUAL")
            elif engine == DbEngine.POSTGRES:
                sel_stmt = ("SELECT * FROM public.pg_get_table_def("
                            f"'{table_schema.lower()}', '{table_name.lower()}', false)")
            elif engine == DbEngine.SQLSERVER:
                # sel_stmt = f"EXEC sp_help '{schema_name}.{table_name}'"
                sel_stmt = ("SELECT OBJECT_DEFINITION (OBJECT_ID("
                            f"'{table_schema.lower()}.{table_name.upper()}'))")

            # execute the query
            recs: list[tuple[str]] = db_select(sel_stmt=sel_stmt,
                                               engine=engine,
                                               connection=connection,
                                               committable=committable,
                                               errors=errors)
            # process the query result
            if isinstance(recs, list) and recs:
                result = recs[0][0].strip()
        else:
            # 'table_name' not schema-qualified, report the problem
            errors.append(f"Table '{table_name}' not properly schema-qualified")

    return result


def db_get_column_metadata(table_name: str,
                           column_name: str,
                           engine: DbEngine = None,
                           connection: Any = None,
                           committable: bool = None,
                           errors: list[str] = None) -> tuple[str, int, int, bool] | None:
    """
    Retrieve metadata on column *column_name* in *table_name*.

    The metadata is returned as a tuple containing:
        - data type: the 'udt_name' (name of the data type used for the column)
        - precision:
            -  numeric types: total number of digits
            - integer types: size in bytes
            -  char/varchar types: maximum number of characters
            - float types: number of bits used to store the mantissa (significant digits) - 24 (float4) or 53 (float8)
            - timestamp types: NULL
        - scale:
            - numeric types: number of decimal digits
            - integer types: 0
            - all other types: NULL
        - nullability: whether the column can hold NULL values

    The parameter *committable* is relevant only if *connection* is provided, and is otherwise ignored.
    A rollback is always attempted, if an error occurs.

    :param table_name: the, possibly schema-qualified, name of the table
    :param column_name: the name of the column to retrieve
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit operation on errorless completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: the metadata on column *column_name* in *table_name*, or *None* if error or metadata not found
    """
    # initialize the return variable
    result: tuple[str, int, int, bool] | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    if engine:
        # determine the table schema
        table_schema: str
        table_schema, table_name = table_name.split(sep=".") if "." in table_name else (None, table_name)

        # build the query
        sel_stmt: str | None = None
        where_data: dict[str, str] | None = None
        if engine == DbEngine.MYSQL:
            pass
        if engine == DbEngine.ORACLE:
            sel_stmt = ("SELECT DATA_TYPE, DATA_LENGTH, "
                        "DATA_PRECISION, DATA_SCALE, NULLABLE FROM ALL_TAB_COLUMNS")
            where_data = {
                "TABLE_NAME": table_name.upper(),
                "COLUMN_NAME": column_name.upper()
            }
            if table_schema:
                where_data["OWNER"] = table_schema.upper()
        elif engine == DbEngine.POSTGRES:
            sel_stmt = ("SELECT udt_name, character_maximum_length, "
                        "numeric_precision, numeric_scale, is_nullable FROM information_schema.columns")
            where_data = {
                "table_name": table_name.lower(),
                "column_name": column_name.lower()
            }
            if table_schema:
                where_data["table_schema"] = table_schema.lower()
        elif engine == DbEngine.SQLSERVER:
            sel_stmt = ("SELECT UDT_NAME, CHARACTER_MAXIMUM_LENGTH, "
                        "NUMERIC_PRECISION, NUMERIC_SCALE, IS_NULLABLE FROM information_schema.columns")
            where_data = {
                "TABLE_NAME": table_name.upper(),
                "COLUMN_NAME": column_name.upper()
            }
            if table_schema:
                where_data["TABLE_SCHEMA"] = table_schema.upper()

        # execute the query
        # noinspection PyTypeChecker
        recs: list[tuple[str, int, int, int, str]] = db_select(sel_stmt=sel_stmt,
                                                               where_data=where_data,
                                                               engine=engine,
                                                               connection=connection,
                                                               committable=committable,
                                                               errors=errors)
        # process the query result
        if isinstance(recs, list) and recs:
            rec = recs[0]
            result = (rec[0], rec[1 if "char" in rec[0] else 2], rec[3], rec[4].startswith("Y"))

    return result


def db_get_columns_metadata(table_name: str,
                            engine: DbEngine = None,
                            connection: Any = None,
                            committable: bool = None,
                            errors: list[str] = None) -> list[tuple[str, str, int, int, bool]] | None:
    """
    Retrieve metadata on all columns in *table_name*.

    The metadata is returned as a list of tuples, each one containing:
        - column name: the name of the column
        - data type: the 'udt_name' (name of the data type used for the column)
        - precision:
            -  numeric types: total number of digits
            - integer types: size in bytes
            -  char/varchar types: maximum number of characters
            - float types: number of bits used to store the mantissa (significant digits) - 24 (float4) or 53 (float8)
            - timestamp types: NULL
        - scale:
            - numeric types: number of decimal digits
            - integer types: 0
            - all other types: NULL
        - nullability: whether the column can hold NULL values

    The parameter *committable* is relevant only if *connection* is provided, and is otherwise ignored.
    A rollback is always attempted, if an error occurs.

    :param table_name: the, possibly schema-qualified, name of the table
    :param engine: the database engine to use (uses the default engine, if not provided)
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit operation on errorless completion
    :param errors: incidental error messages (might be a non-empty list)
    :return: a list of tuples with the metadata on all columns in *table_name*, or *None* if error
    """
    # initialize the return variable
    result: list[tuple[str, str, int, int, bool]] | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine,
                            errors=errors)
    if engine:
        # determine the table schema
        table_schema: str
        table_schema, table_name = table_name.split(sep=".") if "." in table_name else (None, table_name)

        # build the query
        sel_stmt: str | None = None
        where_data: dict[str, str] | None = None
        if engine == DbEngine.MYSQL:
            pass
        if engine == DbEngine.ORACLE:
            sel_stmt = ("SELECT COLUMN_NAME, DATA_TYPE, DATA_LENGTH, "
                        "DATA_PRECISION, DATA_SCALE, NULLABLE FROM ALL_TAB_COLUMNS")
            where_data = {
                "TABLE_NAME": table_name.upper()
            }
            if table_schema:
                where_data["OWNER"] = table_schema.upper()
        elif engine == DbEngine.POSTGRES:
            sel_stmt = ("SELECT column_name, udt_name, character_maximum_length, "
                        "numeric_precision, numeric_scale, is_nullable FROM information_schema.columns")
            where_data = {
                "table_name": table_name.lower()
            }
            if table_schema:
                where_data["table_schema"] = table_schema.lower()
        elif engine == DbEngine.SQLSERVER:
            sel_stmt = ("SELECT COLUMN_NAME, UDT_NAME, CHARACTER_MAXIMUM_LENGTH, "
                        "NUMERIC_PRECISION, NUMERIC_SCALE, IS_NULLABLE FROM information_schema.columns")
            where_data = {
                "TABLE_NAME": table_name.upper()
            }
            if table_schema:
                where_data["TABLE_SCHEMA"] = table_schema.upper()

        # execute the query
        # noinspection PyTypeChecker
        recs: list[tuple[str, str, int, int, int, str]] = db_select(sel_stmt=sel_stmt,
                                                                    where_data=where_data,
                                                                    engine=engine,
                                                                    connection=connection,
                                                                    committable=committable,
                                                                    errors=errors)
        # process the query result
        if isinstance(recs, list) and recs:
            result = [(rec[0], rec[1], rec[2 if "char" in rec[1] else 3], rec[4], rec[5].startswith("Y"))
                      for rec in recs]
    return result



from __future__ import annotations  # allow forward references
from collections.abc import Callable
from contextlib import suppress
from datetime import datetime, UTC
from enum import StrEnum, auto
from logging import Logger
from pypomes_core import (
    APP_PREFIX, env_get_int, env_get_bool, obj_positional, str_sanitize
)
from sys import getrefcount
from time import sleep
from threading import Lock
from typing import Any, Final

from .db_common import _DB_LOGGERS, DbEngine, _assert_engine


class DbPoolEvent(StrEnum):
    """
    Pool events for tuning and monitoring.
    """
    CREATE = auto()
    CHECKOUT = auto()
    CHECKIN = auto()
    CLOSE = auto()


class _PoolParam(StrEnum):
    """
    Parameters for configuring the connection pool.
    """
    SIZE = auto()
    TIMEOUT = auto()
    RECYCLE = auto()
    VERIFY = auto()


class _ConnStage(StrEnum):
    """
    Stage data of connections in pool.
    """
    CONNECTION = auto()                 # the connection object
    TIMESTAMP = auto()                  # creation datetime (POSIX timestamp)
    AVAILABLE = auto()                  # status (True: available; False: taken)


# available DbConnectionPool instances:
# {
#    <DbEngine>: <DbConnectionPool>,
#    ...
# }
_POOL_INSTANCES: Final[dict[DbEngine, DbConnectionPool]] = {}

# pool instances access lock
_instances_lock: Final[Lock] = Lock()


def __get_pool_params() -> dict[DbEngine, dict[_PoolParam, Any]]:
    """
    Establish the connection pool parameters for supported databases, from values in environment variables.

    To specify database connection parameters with environment variables, use the set:
      - *<APP_PREFIX>_DB_POOL_SIZE*: maximum number of connections in the pool (defaults to 40 connections)
      - *<APP_PREFIX>_DB_POOL_TIMEOUT*: number of seconds to wait for a connection to become available,
        before failing the request (defaults to 60 seconds)
      - *<APP_PREFIX>_DB_POOL_RECYCLE*: number of seconds after which connections in the pool are closed
        (defaults to 3600 seconds)
      - *<APP_PREFIX>_DB_POOL_VERIFY*: ensures the connection is still active and healthy before it is
        checked out from the pool, by executing a lightweight query - if the verification fails,
        the connection is discarded (defaults to *True*)

    These variables apply to any of the supported database engines. To specify values for a specific engine,
    replace *_DB_* with *_MSQL_*, *_ORCL_*, *_PG_*, or *_SQLS_*, respectively for *mysql*, *oracle*,
    *postgres*, and *sqlserver*.

    :return: the connection parameters for supported database engines
    """
    # initialize the return variable
    result: dict[DbEngine, dict[_PoolParam, Any]] = {}

    for engine in DbEngine:
        prefix: str = obj_positional(engine,
                                     keys=tuple(DbEngine),
                                     values=("MSQL", "ORCL", "PG", "SQLS"))
        # establish pool size
        size: int = env_get_int(key=f"{APP_PREFIX}_{prefix}_POOL_SIZE")
        if not isinstance(size, int) or size < 0:
            size = env_get_int(key=f"{APP_PREFIX}_DB_POOL_SIZE",
                               def_value=40)
        # establish pool timeout
        timeout: int = env_get_int(key=f"{APP_PREFIX}_{prefix}_POOL_TIMEOUT")
        if not timeout:
            timeout = env_get_int(key=f"{APP_PREFIX}_DB_POOL_TIMEOUT",
                                  def_value=60)
        # establish pool recycle
        recycle: int = env_get_int(key=f"{APP_PREFIX}_{prefix}_POOL_RECYCLE")
        if not recycle:
            recycle = env_get_int(key=f"{APP_PREFIX}_DB_POOL_RECYCLE",
                                  def_value=3600)
        # establish connection verification
        verify: bool = env_get_bool(key=f"{APP_PREFIX}_{prefix}_POOL_VERIFY")
        if verify is None:
            verify = env_get_bool(key=f"{APP_PREFIX}_DB_POOL_VERIFY",
                                  def_value=True)
        result[engine] = {
            _PoolParam.SIZE: size,
            _PoolParam.TIMEOUT: timeout,
            _PoolParam.RECYCLE: recycle,
            _PoolParam.VERIFY: verify
        }

    return result


# DbConnectionPool instance params:
# {
#    <DbEngine>: {
#       <_PoolParam.RECYCLE>: <int>,
#       <_PoolParam.SIZE>: <int>,
#       <_PoolParam.TIMEOUT>: <int>,
#       <_PoolParam.VERIFY>: <bool>
#    },
#    ...
# }
_POOL_PARAMS: Final[dict[DbEngine, dict[_PoolParam, Any]]] = __get_pool_params()


class DbConnectionPool:
    """
    A robust, transparent, and efficient implementation of a database connection pool.

    This is a mechanism to manage database connections efficiently by reusing them, instead of their being
    repeatedly created and closed. As a result, overall performance is improved, latency is reduced, and
    resource usage is optimized, especially in applications with frequent database interactions.
    The connections are lazily created, as demand for connections arise and cannot be met by the stock
    already in the pool.

    This implementation follows the best practices in the industry, mirroring state-of-the-art products
    such as the *SQLAlchemy* pool package. These are the configuration parameters:
      - *pool_size*: maximum number of connections in the pool
      - *pool_timeout*: number of seconds to wait for a connection to become available, before failing the request
      - *pool_recycle*: number of seconds after which connections in the pool are closed
      - *pool_verify*: whether to ensure that the connection is still healthy before it is checked out from the pool

    These are the events that a pool client may hook up to, with a call to *on_event_actions()*, by
    specifying a callback function to be invoked, and/or SQL commands to be executed:
      - *create*: when a connection is created, allowing for connection session settings customization
      - *checkout*: when a connection is retrieved from the pool, allowing for fine-tuning the connection settings
      - *checkin*: when a connection is returned to the pool, allowing for cleanup of session state
      - *close*: when a connection is closed, allowing for resources cleanup and connection life cycle auditing

    The modules handling the native database operations (namely, *mysql_pomes.py*, *oracle_pomes.py*,
    *postgres_pomes.py*, and *sqlserver_pomes.py*) have no awareness of this pool. The higher level module
    *db_pomes.py* will attempt to obtain a connection from this pool on *db_connect()*, and to return it
    on *db_close()*. It is worth emphasizing that a *close()* operation should not be invoked directly
    on the native connection, as this would prevent it from being reclaimed, and cause it to be eventually
    discarded, thus defeating the very purpose of the pool.

    Finally, this implementation does not appy to Goggle's *spanner* database engine, as in that case a
    built-in connection pool is always handled under the hood, thus rendering this implementation as unnecessary.
    """
    def __init__(self,
                 engine: DbEngine = None,
                 /,
                 *,
                 pool_size: int = None,
                 pool_timeout: int = None,
                 pool_recycle: int = None,
                 pool_verify: bool = None,
                 errors: list[str] = None) -> None:
        """
        Construct a connection pool specific for the database provided in *engine*.

        The database engine specified must have been previously configured. This pool implementation does not
        apply to Google's *spanner* database engine, as it handles a built-in connection pool under the hood.

        If not provided, the values for pool configuration parameters specified in the environment variables
        suffixed with their uppercase names (namely, *POOL_SIZE*, *POOL_TIMEOUT*, *POOL_RECYCLE*,
        and *POOL_VERIFY*) are used. If still not specified, these are the default values used:
          - *pool_size*: 40 connections
          - *pool_timeout*: 60 seconds
          - *pool_recycle*: 3600 seconds (60 minutos)
          - *pool_verify*: True

        All instance attributes are final and should not be directly changed, lest the pool malfunction.

        :param engine: the database engine to use (uses the default engine, if not provided)
        :param pool_size: maximum number of connections to keep in the pool
        :param pool_timeout: number of seconds to wait for an available connection before failing
        :param pool_recycle: number of seconds after which connections in the pool are closed and reopened
        :param pool_verify: whether to ensure that the connection is still healthy before it is checked out
        :param errors: incidental error messages (might be a non-empty list)
        """
        # declare the instance variables
        self.db_engine: DbEngine
        self.pool_size: int
        self.pool_timeout: int
        self.pool_recycle: int
        self.pool_verify: bool
        self.stage_lock: Lock
        self.event_lock: Lock
        self.event_callbacks: dict[DbPoolEvent, tuple[Callable[[Any, Logger], None], list[str]]]
        self.conn_data: list[dict[_ConnStage, Any]]

        # make sure a configured database engine has been specified
        engine: DbEngine = _assert_engine(engine=engine,
                                          errors=errors)

        # obtain the default values for the pool parameters
        pool_params: dict[_PoolParam, Any] | None = None
        if engine:
            msg: str | None = None
            if engine == DbEngine.SPANNER:
                msg = f"pool does not apply to {engine}"
            else:
                with _instances_lock:
                    if engine in _POOL_INSTANCES:
                        msg += f"{engine} pool already exists"
                    else:
                        pool_params = _POOL_PARAMS[engine]
            if msg:
                msg = "Attempt to create connection pool failed: " + msg
                if _DB_LOGGERS[engine]:
                    _DB_LOGGERS[engine].error(msg)
                if isinstance(errors, list):
                    errors.append(msg)

        if pool_params:
            if not isinstance(pool_size, int):
                pool_size = pool_params[_PoolParam.SIZE]
            if pool_size > 0:
                self.db_engine = engine
                self.pool_size = pool_size
                self.pool_timeout = pool_timeout \
                    if isinstance(pool_timeout, int) else pool_params[_PoolParam.TIMEOUT]
                self.pool_recycle = pool_recycle \
                    if isinstance(pool_recycle, int) else pool_params[_PoolParam.RECYCLE]
                self.pool_verify = pool_verify \
                    if isinstance(pool_verify, bool) else pool_params[_PoolParam.VERIFY]
                self.stage_lock = Lock()
                self.event_lock = Lock()

                self.event_callbacks = {DbPoolEvent.CREATE: (None, []),
                                        DbPoolEvent.CHECKOUT: (None, []),
                                        DbPoolEvent.CHECKIN: (None, []),
                                        DbPoolEvent.CLOSE: (None, [])}
                self.conn_data = []

                # register this instance
                _POOL_INSTANCES[engine] = self
                if _DB_LOGGERS[engine]:
                    msg: str = (f"{self.db_engine} pool created: size {self.pool_size}, "
                                f"timeout {self.pool_timeout}, recycle {self.pool_recycle}")
                    _DB_LOGGERS[engine].info(msg=msg)
            else:
                msg: str = f"{engine} pool not created: specified size was {pool_size}"
                if _DB_LOGGERS[engine]:
                    _DB_LOGGERS[engine].error(msg=msg)
                if isinstance(errors, list):
                    errors.append(msg)

    def connect(self,
                start: float = None,
                errors: list[str] = None) -> Any | None:
        """
        Obtain a pooled connection.

        :param start: start of operation
        :param errors: incidental error messages (might be a non-empty list)
        :return: a connection from the pool, or *None* if error
        """
        # initialize the return variable
        result: Any = None

        # required, lest the state of 'errors' be tested
        curr_errors: list[str] = []

        # make sure to have a start timestamp
        if not isinstance(start, float):
            start = datetime.now(tz=UTC).timestamp()

        with self.stage_lock:
            while not curr_errors and not result:
                reclaimed: bool = False
                # traverse the connection data in reverse order
                for i in range(len(self.conn_data) - 1, -1, -1):
                    conn: Any = self.conn_data[i][_ConnStage.CONNECTION]

                    # connection is available
                    if self.conn_data[i][_ConnStage.AVAILABLE]:
                        # connection exhausted its lifetime
                        if datetime.now(tz=UTC).timestamp() > \
                                self.conn_data[i][_ConnStage.TIMESTAMP] + self.pool_recycle:
                            # close the exhausted connection
                            self.__act_on_event(event=DbPoolEvent.CLOSE,
                                                conn=conn)
                            with suppress(Exception):
                                conn.close()
                            # dispose of exhausted connection
                            self.conn_data.pop(i)
                            if _DB_LOGGERS[self.db_engine]:
                                msg: str = (f"The exhausted connection {id(conn)} was removed "
                                            f"from the {self.db_engine} pool, count = {len(self.conn_data)}")
                                _DB_LOGGERS[self.db_engine].debug(msg=msg)
                        # assess the connection
                        else:
                            if self.pool_verify:
                                # assert connection health
                                stmt = "SELECT 1"
                                if self.db_engine == DbEngine.ORACLE:
                                    stmt += " FROM DUAL"
                                try:
                                    cursor: Any = conn.cursor()
                                    # parameter name:
                                    #   - mysql-connector-python (mysql): none
                                    #   - oracledb (oracle): 'statement'
                                    #   - psycopg2 (postgres): 'query'
                                    #   - pyodbc (sqlserver): 'sql'
                                    cursor.execute(stmt)
                                    cursor.fetchone()
                                    cursor.close()
                                except Exception as e:
                                    # dispose of bad connection
                                    self.conn_data.pop(i)
                                    conn = None
                                    if _DB_LOGGERS[self.db_engine]:
                                        msg: str = (f"Connection {id(conn)} assessment "
                                                    f"failed: {str_sanitize(f'{e}')}")
                                        _DB_LOGGERS[self.db_engine].warning(msg=msg)
                                        msg = (f"The bad connection {id(conn)} was removed from "
                                               f"the {self.db_engine} pool, count = {len(self.conn_data)}")
                                        _DB_LOGGERS[self.db_engine].debug(msg=msg)
                            if conn:
                                # retrieve the connection, and mark it as taken
                                result = conn
                                self.conn_data[i][_ConnStage.AVAILABLE] = False
                                break

                    # connection was closed elsewhere
                    elif hasattr(conn, "closed") and conn.closed:
                        # dispose of closed connection
                        self.conn_data.pop(i)
                        if _DB_LOGGERS[self.db_engine]:
                            msg: str = (f"The closed connection {id(conn)} was removed from "
                                        f"the {self.db_engine} pool, count = {len(self.conn_data)}")
                            _DB_LOGGERS[self.db_engine].debug(msg=msg)

                    # connection is no longer in use - these are the 3 remaining references:
                    #   - pool repository 'self.conn_data[i][_ConnStage.CONNECTION]'
                    #   - local variable 'conn'
                    #   - argument to 'getrefcount()'
                    elif getrefcount(conn) < 4:
                        # reclaim the connection
                        self.__act_on_event(event=DbPoolEvent.CHECKIN,
                                            conn=conn)
                        # mark the connection as available
                        self.conn_data[i][_ConnStage.AVAILABLE] = True
                        reclaimed = True
                        if _DB_LOGGERS[self.db_engine]:
                            msg: str = (f"The stray connection {id(conn)} "
                                        f"was reclaimed by the {self.db_engine} pool")
                            _DB_LOGGERS[self.db_engine].debug(msg=msg)
                # end of conn data traversal
                if not result and not reclaimed:
                    # no connection retrieved or reclaimed, obtain a new one
                    if len(self.conn_data) < self.pool_size:
                        match self.db_engine:
                            case DbEngine.MYSQL:
                                from . import mysql_pomes
                                conn = mysql_pomes.connect(autocommit=False,
                                                           errors=curr_errors)
                            case DbEngine.ORACLE:
                                from . import oracle_pomes
                                conn = oracle_pomes.connect(autocommit=False,
                                                            errors=curr_errors)
                            case DbEngine.POSTGRES:
                                from . import postgres_pomes
                                conn = postgres_pomes.connect(autocommit=False,
                                                              errors=curr_errors)
                            case DbEngine.SQLSERVER:
                                from . import sqlserver_pomes
                                conn = sqlserver_pomes.connect(autocommit=False,
                                                               errors=curr_errors)
                        if not curr_errors:
                            self.__act_on_event(event=DbPoolEvent.CREATE,
                                                conn=conn)
                            # retrieve and store the connection
                            result = conn
                            self.conn_data.append({
                                _ConnStage.AVAILABLE: False,
                                _ConnStage.CONNECTION: result,
                                _ConnStage.TIMESTAMP: datetime.now(tz=UTC).timestamp()
                            })
                            if _DB_LOGGERS[self.db_engine]:
                                _DB_LOGGERS[self.db_engine].debug(msg=f"Connection {id(result)} "
                                                                      f"created by the {self.db_engine} pool, "
                                                                      f"count = {len(self.conn_data)}")
                    else:
                        # connection pool is at capacity
                        break
        if not curr_errors:
            if result:
                # connection retrieved
                self.__act_on_event(event=DbPoolEvent.CHECKOUT,
                                    conn=result)
                if _DB_LOGGERS[self.db_engine]:
                    _DB_LOGGERS[self.db_engine].debug(msg=f"Connection {id(result)} "
                                                          f"delivered by the {self.db_engine} pool")
            elif datetime.now(tz=UTC).timestamp() < start + self.pool_timeout:
                # no connection retrieved, there is still time to retry
                sleep(1.5)
                result = self.connect(start=start,
                                      errors=curr_errors)
            else:
                curr_errors.append("Timeout waiting for available connection")

        if curr_errors and isinstance(errors, list):
            errors.extend(curr_errors)

        return result

    def on_event_actions(self,
                         event: DbPoolEvent,
                         callback: Callable[[Any, Logger], None] = None,
                         stmts: list[str] = None) -> None:
        """
        Specify a callback function to be invoked, and/or SQL commands to be executed, when *event* occurs.

        The possible events are:
          - *PoolEvent.CREATE*: a connection is created and stored in the pool
          - *PoolEvent.CHECKOUT*: a connection is retrieved from the pool
          - *PoolEvent.CHECKIN*: a connection is returned to the pool
          - *PoolEven.CLOSE*: a connection is closed and removed from the pool

        If *callback* is not specified, any current callback is removed for *event*, otherwise it will be
        invoked with the following parameters:
          - *connection*: the native connection obtained from the underlying database driver
          - *logger*: the logger in use by the operation that raised the event (may be *None*)

        If *stmts* is not specified, any current statements are removed for *event*, otherwise it holds
        one or more SQL commands to be executed, aiming to, among others, and depending on the event:
          - initialize, set, reset, or cleanup the session state
          - set encoding, timezone, and sort order
          - verify connection health
          - audit connection life cycle
          - cleanup resources
          - log usage to database

        :param event: the reference event
        :param callback: the function to be invoked
        :param stmts: optional list of SQL commands to be executed
        """
        with self.event_lock:
            # register the event hook and the associated SQL commands
            self.event_callbacks[event] = (callback, stmts)

    def reclaim(self,
                connection: Any) -> bool:
        """
        Reclaim the connection given in *conn*, allowing for its reuse.

        :param connection: the connection to be reclaimed
        :return: *True* if *conn* was successfully reclaimed, *False* otherwise
        """
        # initialize the return variable
        result: bool = False

        with self.stage_lock:
            for data in self.conn_data:
                if data[_ConnStage.CONNECTION] == connection:
                    self.__act_on_event(event=DbPoolEvent.CHECKIN,
                                        conn=connection)
                    # mark the connection as available
                    data[_ConnStage.AVAILABLE] = True
                    result = True
                    if _DB_LOGGERS[self.db_engine]:
                        _DB_LOGGERS[self.db_engine].debug(msg=f"Connection {id(connection)} "
                                                              f"returned to the {self.db_engine} pool")
                    break

        return result

    def terminate(self,
                  logger: Logger = None) -> None:
        """
        Terminate the pool, releasing all held resources.

        Upon termination, all references to this instance should be disposed of,
        as invocation of any of its functions is guaranteed to fail.
        """
        with _instances_lock:
            _POOL_INSTANCES.pop(self.db_engine)

        with self.stage_lock:
            self.conn_data.clear()

        with self.event_lock:
            self.event_callbacks.clear()

        if logger:
            logger.debug(msg=f"{self.db_engine} pool terminated")

    def __act_on_event(self,
                       event: DbPoolEvent,
                       conn: Any) -> None:
        """
        Act on *event*, by invoking the hooked callback function, and/or executing the registered SQL statements.

        The possible events are:
          - *PoolEvent.CREATE*: a connection is created and stored in the pool
          - *PoolEvent.CHECKOUT*: a connection is retrieved from the pool
          - *PoolEvent.CHECKIN*: a connection is returned to the pool
          - *PoolEven.CLOSE*: a connection is closed and removed from the pool

        If a callback function has been hooked to *event*, it is invoked with the following parameters:
          - the native connection obtained from the underlying database driver, its type being one of
              - *mysql.connector.aio.MySQLConnectionAbstract* (mysql-connector-python)
              - *oracle.Connection* (oracledb)
              - *psycopg2._psycopg.connection* (psycopg2)
              - *pyodbc.connection* (pyodbc)
          - the logger in use by the operation that raised the event (may be *None*)

        :param event: the reference event
        :param conn: the reference connection
        """
        with self.event_lock:
            # invoke the callback
            # noinspection PyTypeChecker
            event_callback: Callable[[Any, Logger], None] = self.event_callbacks[event][0]
            if callable(event_callback):
                # noinspection PyCallingNonCallable
                event_callback(conn,
                               _DB_LOGGERS[self.db_engine])
                if _DB_LOGGERS[self.db_engine]:
                    msg: str = (f"Event '{event}' on {self.db_engine}, "
                                f"connection {id(conn)}: '{event_callback.__name__}' invoked")
                    _DB_LOGGERS[self.db_engine].debug(msg=msg)
            # execute the statements
            stmts: list[str] = self.event_callbacks[event][1]
            from . import db_pomes
            for stmt in stmts:
                errors: list[str] = []
                db_pomes.db_execute(exc_stmt=stmt,
                                    engine=self.db_engine,
                                    connection=conn,
                                    errors=errors)
                if _DB_LOGGERS[self.db_engine] and not errors:
                    msg: str = (f"Event '{event}' on {self.db_engine}, "
                                f"connection {id(conn)}: '{stmt}' executed")
                    _DB_LOGGERS[self.db_engine].debug(msg=msg)


def db_get_pool(engine: DbEngine = None) -> DbConnectionPool | None:
    """
    Retrieve the instance of the connection pool associated with *engine*.

    :param engine: the database engine to use (uses the default engine, if not provided)
    :return: the connection pool for *engine*, or *None* if no connection has been retrieved
    """
    # initialize the return variable
    result: DbConnectionPool | None = None

    # assert the database engine
    engine = _assert_engine(engine=engine)
    if engine:
        with _instances_lock:
            result = _POOL_INSTANCES.get(engine)

    return result


def db_pool_acquire(engine: DbEngine = None) -> Any | None:
    """
    Obtain a pooled connection for *engine*.

    :param engine: the database engine to use (uses the default engine, if not provided)
    :return: a pooled connection to *engine*, or *None* if it unknown or has not been configured
    """
    # initialize the return variable
    result: Any = None

    # assert the database engine
    engine = _assert_engine(engine=engine)
    if engine:
        pool: DbConnectionPool = db_get_pool(engine=engine)
        # obtain the pooled connection
        if pool:
            result = pool.connect()

    return result


def db_pool_release(connection: Any,
                    engine: DbEngine = None) -> bool:
    """
    Return *conn* to the pool, to allow it to be reused.

    :param connection: the connection to return to the pool
    :param engine: the database engine to use (uses the default engine, if not provided)
    :return: *True* if *conn* was successfully reclaimed by the pool, *False* otherwise
    """
    # initialize the return variable
    result: bool = False

    # assert the database engine
    engine = _assert_engine(engine=engine)
    if engine:
        # reclaim the connection
        pool: DbConnectionPool = db_get_pool(engine=engine)
        if pool:
            result = pool.reclaim(connection=connection)

    return result



from datetime import date, datetime
from mysql import connector
from mysql.connector.aio import MySQLConnectionAbstract
from pypomes_core import DateFormat, DatetimeFormat
from typing import Any, Final

from .db_common import (
    _DB_LOGGERS, DbEngine, DbParam, _get_params, _except_msg
)

RESERVED_WORDS: Final[list[str]] = [
    "ACCESSIBLE", "ADD", "ALL", "ALTER", "ANALYZE", "AND", "AS", "ASC", "ASENSITIVE", "BEFORE", "BETWEEN",
    "BIGINT", "BINARY", "BLOB", "BOTH", "BY", "CALL", "CASCADE", "CASE", "CHANGE", "CHAR", "CHARACTER",
    "CHECK", "COLLATE", "COLUMN", "CONDITION", "CONSTRAINT", "CONTINUE", "CONVERT", "CREATE", "CROSS",
    "CUBE", "CUME_DIST", "CURRENT_DATE", "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT_USER", "CURSOR",
    "DATABASE", "DATABASES", "DAY_HOUR", "DAY_MICROSECOND", "DAY_MINUTE", "DAY_SECOND", "DEC", "DECIMAL",
    "DECLARE", "DEFAULT", "DELAYED", "DELETE", "DENSE_RANK", "DESC", "DESCRIBE", "DETERMINISTIC", "DISTINCT",
    "DISTINCTROW", "DIV", "DOUBLE", "DROP", "DUAL", "EACH", "ELSE", "ELSEIF", "EMPTY", "ENCLOSED", "ESCAPED",
    "EXCEPT", "EXECUTE", "EXISTS", "EXIT", "EXPLAIN", "FALSE", "FETCH", "FIRST_VALUE", "FLOAT", "FLOAT4", "FLOAT8",
    "FOR", "FORCE", "FOREIGN", "FROM", "FULLTEXT", "FUNCTION", "GENERATED", "GET", "GRANT", "GROUP", "GROUPING",
    "GROUPS", "HAVING", "HIGH_PRIORITY", "HOUR_MICROSECOND", "HOUR_MINUTE", "HOUR_SECOND", "IF", "IGNORE", "ILIKE",
    "IN", "INDEX", "INFILE", "INNER", "INOUT", "INSENSITIVE", "INSERT", "INT", "INT1", "INT2", "INT3", "INT4", "INT8",
    "INTEGER", "INTERVAL", "INTO", "IO_AFTER_GTIDS", "IO_BEFORE_GTIDS", "IS", "ITERATE", "JOIN", "JSON_TABLE",
    "KEY", "KEYS", "KILL", "LAG", "LAST_VALUE", "LATERAL", "LEAD", "LEADING", "LEAVE", "LEFT", "LIKE", "LIMIT",
    "LINEAR", "LINES", "LOAD", "LOCALTIME", "LOCALTIMESTAMP", "LOCK", "LONG", "LONGBLOB", "LONGTEXT", "LOOP",
    "LOW_PRIORITY", "MASTER_BIND", "MASTER_SSL_VERIFY_SERVER_CERT", "MATCH", "MAXVALUE", "MEDIUMBLOB", "MEDIUMINT",
    "MEDIUMTEXT", "MIDDLEINT", "MINUTE_MICROSECOND", "MINUTE_SECOND", "MOD", "MODIFIES", "NATURAL", "NOT",
    "NO_WRITE_TO_BINLOG", "NTH_VALUE", "NTILE", "NULL", "NUMERIC", "OF", "ON", "OPTIMIZE", "OPTIMIZER_COSTS",
    "OPTION", "OPTIONALLY", "OR", "ORDER", "OUT", "OUTER", "OUTFILE", "OVER", "PARTITION", "PERCENT_RANK",
    "PRECISION", "PRIMARY", "PROCEDURE", "PURGE", "RANGE", "RANK", "READ", "READS", "READ_WRITE", "REAL",
    "RECURSIVE", "REF_SYSTEM_ID", "REFERENCES", "REGEXP", "RELEASE", "RENAME", "REPEAT", "REPLACE", "REQUIRE",
    "RESIGNAL", "RESTRICT", "RETURN", "REVOKE", "RIGHT", "RLIKE", "ROW", "ROWS", "ROW_NUMBER", "SCHEMA",
    "SCHEMAS", "SECOND_MICROSECOND", "SELECT", "SENSITIVE", "SEPARATOR", "SET", "SHOW", "SIGNAL", "SMALLINT",
    "SPATIAL", "SPECIFIC", "SQL", "SQLEXCEPTION", "SQLSTATE", "SQLWARNING", "SQL_BIG_RESULT",
    "SQL_CALC_FOUND_ROWS", "SQL_SMALL_RESULT", "SSL", "STARTING", "STORED", "STRAIGHT_JOIN", "SYSTEM",
    "TABLE", "TERMINATED", "THEN", "TINYBLOB", "TINYINT", "TINYTEXT", "TO", "TRAILING", "TRIGGER", "TRUE",
    "UNDO", "UNION", "UNIQUE", "UNLOCK", "UNSIGNED", "UPDATE", "USAGE", "USE", "USING", "UTC_DATE", "UTC_TIME",
    "UTC_TIMESTAMP", "VALUES", "VARBINARY", "VARCHAR", "VARCHARACTER", "VARYING", "VIRTUAL", "WHEN", "WHERE",
    "WHILE", "WINDOW", "WITH", "WRITE", "XOR", "YEAR_MONTH", "ZEROFILL"
]


def connect(autocommit: bool = None,
            errors: list[str] = None) -> MySQLConnectionAbstract | None:
    """
    Obtain and return a connection to the database.

    Return *None* if the connection could not be obtained.

    :param autocommit: whether the connection is to be in autocommit mode (defaults to *False*)
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the connection to the database, or *None* if error
    """
    # initialize the return variable
    result: MySQLConnectionAbstract | None = None

    # retrieve the connection parameters
    db_params: dict[DbParam, Any] = _get_params(DbEngine.MYSQL)

    # obtain a connection to the database
    try:
        result = connector.connect(database=db_params.get(DbParam.NAME),
                                   host=db_params.get(DbParam.HOST),
                                   port=db_params.get(DbParam.PORT),
                                   user=db_params.get(DbParam.USER),
                                   password=db_params.get(DbParam.PWD))
        # establish the connection's autocommit mode
        result.autocommit = isinstance(autocommit, bool) and autocommit
    except Exception as e:
        msg: str = _except_msg(exception=e,
                               connection=result,
                               engine=DbEngine.ORACLE)
        if _DB_LOGGERS[DbEngine.MYSQL]:
            _DB_LOGGERS[DbEngine.MYSQL].error(msg=msg)
        if isinstance(errors, list):
            errors.append(msg)

    # log errors
    if _DB_LOGGERS[DbEngine.MYSQL] and errors:
        _DB_LOGGERS[DbEngine.MYSQL].error(msg=f"Error connecting to '{db_params.get(DbParam.NAME)} '"
                                              f"at '{db_params.get(DbParam.NAME)}'")
    return result


def bind_arguments(stmt: str,
                   bind_vals: list[Any]) -> str:
    """
    Replace the placeholders in *query_stmt* with the values in *bind_vals*, and return the modified query statement.

    Note that using a statement in a situation where values for types other than *bool*, *str*, *int*, *float*,
    *date*, or *datetime* were replaced, may bring about undesirable consequences, as the standard string
    representations for these other types would be used.

    :param stmt: the query statement
    :param bind_vals: the values to replace the placeholders with
    :return: the query statement with the placeholders replaced with their corresponding values
    """
    # initialize the return variable
    result: str = stmt

    # bind the arguments
    for bind_val in bind_vals:
        val: str
        if isinstance(bind_val, bool):
            val = "1" if bind_val else "0"
        elif isinstance(bind_val, int | float):
            val = f"{bind_val}"
        elif isinstance(bind_val, date):
            val = f"STR_TO_DATE('{bind_val.strftime(format=DateFormat.INV)}', '%Y-%m-%d')"
        elif isinstance(bind_val, datetime):
            val = f"STR_TO_DATE('{bind_val.strftime(format=DatetimeFormat.INV)}', '%Y-%m-%d %H:%i:%s')"
        else:
            val = f"'{bind_val}'"
        result = result.replace("?", val, 1)

    return result
